# This file is part of GridCal.
#
# GridCal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GridCal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GridCal.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import pandas as pd
from GridCal.Engine.Simulations.result_types import ResultTypes
from GridCal.Engine.Simulations.results_model import ResultsModel


class NumericPowerFlowResults:

    def __init__(self, V, converged, norm_f, Scalc, ma=None, theta=None, Beq=None, iterations=0, elapsed=0):
        """

        :param V:
        :param converged:
        :param norm_f:
        :param Scalc:
        :param ma:
        :param theta:
        :param Beq:
        :param iterations:
        :param elapsed:
        """
        self.V = V
        self.converged = converged
        self.norm_f = norm_f
        self.Scalc = Scalc
        self.ma = ma
        self.theta = theta
        self.Beq = Beq
        self.iterations = iterations
        self.elapsed = elapsed


class PowerFlowResults:
    """
    A **PowerFlowResults** object is create as an attribute of the
    :ref:`PowerFlowMP<pf_mp>` (as PowerFlowMP.results) when the power flow is run. It
    provides access to the simulation results through its class attributes.

    Attributes:

        **Sbus** (list): Power at each bus in complex per unit

        **voltage** (list): Voltage at each bus in complex per unit

        **Sf** (list): Power through each branch in complex MVA

        **If** (list): Current through each branch in complex per unit

        **loading** (list): Loading of each branch in per unit

        **losses** (list): Losses in each branch in complex MVA

        **tap_module** (list): Computed tap module at each branch in per unit

        **flow_direction** (list): Flow direction at each branch

        **Vbranch** (list): Voltage increment at each branch

        **error** (float): Power flow computed error

        **converged** (bool): Did the power flow converge?

        **Qpv** (list): Reactive power at each PV node in per unit

        **inner_it** (int): Number of inner iterations

        **outer_it** (int): Number of outer iterations

        **elapsed** (float): Simulation duration in seconds

        **methods** (list): Power flow methods used

    """

    def __init__(self, n, m, n_tr, n_hvdc, bus_names, branch_names, transformer_names, hvdc_names, bus_types):

        self.name = 'Power flow'

        self.n = n
        self.m = m
        self.n_tr = n_tr
        self.n_hvdc = n_hvdc

        self.bus_types = bus_types

        self.bus_names = bus_names
        self.branch_names = branch_names
        self.transformer_names = transformer_names
        self.hvdc_names = hvdc_names

        self.Sbus = np.zeros(n, dtype=complex)

        self.voltage = np.zeros(n, dtype=complex)

        self.overvoltage = np.zeros(n, dtype=complex)

        self.undervoltage = np.zeros(n, dtype=complex)

        self.Sf = np.zeros(m, dtype=complex)
        self.St = np.zeros(m, dtype=complex)

        self.If = np.zeros(m, dtype=complex)
        self.It = np.zeros(m, dtype=complex)

        self.ma = np.zeros(m, dtype=float)
        self.theta = np.zeros(m, dtype=float)
        self.Beq = np.zeros(m, dtype=float)

        self.Vbranch = np.zeros(m, dtype=complex)

        self.loading = np.zeros(m, dtype=complex)

        self.flow_direction = np.zeros(m, dtype=float)

        self.transformer_tap_module = np.zeros(n_tr, dtype=float)

        self.losses = np.zeros(m, dtype=complex)

        self.hvdc_losses = np.zeros(self.n_hvdc)

        self.hvdc_Pf = np.zeros(self.n_hvdc)

        self.hvdc_Pt = np.zeros(self.n_hvdc)

        self.hvdc_loading = np.zeros(self.n_hvdc)

        self.plot_bars_limit = 100

        self.convergence_reports = list()

        self.available_results = [ResultTypes.BusVoltageModule,
                                  ResultTypes.BusVoltageAngle,
                                  ResultTypes.BusActivePower,
                                  ResultTypes.BusReactivePower,

                                  ResultTypes.BranchActivePowerFrom,
                                  ResultTypes.BranchReactivePowerFrom,
                                  ResultTypes.BranchActivePowerTo,
                                  ResultTypes.BranchReactivePowerTo,

                                  ResultTypes.BranchActiveCurrentFrom,
                                  ResultTypes.BranchReactiveCurrentFrom,
                                  ResultTypes.BranchActiveCurrentTo,
                                  ResultTypes.BranchReactiveCurrentTo,

                                  ResultTypes.BranchTapModule,
                                  ResultTypes.BranchTapAngle,
                                  ResultTypes.BranchBeq,

                                  ResultTypes.BranchLoading,
                                  ResultTypes.Transformer2WTapModule,
                                  ResultTypes.BranchActiveLosses,
                                  ResultTypes.BranchReactiveLosses,
                                  ResultTypes.BranchVoltage,
                                  ResultTypes.BranchAngles,

                                  ResultTypes.HvdcLosses,
                                  ResultTypes.HvdcPowerFrom,
                                  ResultTypes.HvdcPowerTo]

    @property
    def converged(self):
        """
        Check if converged in all modes
        :return: True / False
        """
        val = True
        for conv in self.convergence_reports:
            val *= conv.converged()
        return val

    @property
    def error(self):
        """
        Check if converged in all modes
        :return: True / False
        """
        val = 0.0
        for conv in self.convergence_reports:
            val = max(val, conv.error())
        return val

    @property
    def elapsed(self):
        """
        Check if converged in all modes
        :return: True / False
        """
        val = 0.0
        for conv in self.convergence_reports:
            val = max(val, conv.elapsed())
        return val

    def copy(self):
        """
        Return a copy of this
        @return:
        """
        val = PowerFlowResults(n=self.n, m=self.m, n_tr=self.n_tr,
                               bus_names=self.bus_names,
                               branch_names=self.branch_names,
                               transformer_names=self.transformer_names)
        val.Sbus = self.Sbus.copy()
        val.voltage = self.voltage.copy()
        val.overvoltage = self.overvoltage.copy()
        val.undervoltage = self.undervoltage.copy()
        val.Sf = self.Sf.copy()
        val.If = self.If.copy()
        val.Vbranch = self.Vbranch.copy()
        val.loading = self.loading.copy()
        val.flow_direction = self.flow_direction.copy()
        val.transformer_tap_module = self.transformer_tap_module.copy()
        val.losses = self.losses.copy()
        val.overloads = self.overloads.copy()

        return val

    def apply_from_island(self, results: "PowerFlowResults", b_idx, br_idx, tr_idx):
        """
        Apply results from another island circuit to the circuit results represented
        here.

        Arguments:

            **results**: PowerFlowResults

            **b_idx**: bus original indices

            **elm_idx**: branch original indices
        """
        self.Sbus[b_idx] = results.Sbus

        self.voltage[b_idx] = results.voltage

        self.Sf[br_idx] = results.Sf

        self.St[br_idx] = results.St

        self.If[br_idx] = results.If

        self.Vbranch[br_idx] = results.Vbranch

        self.loading[br_idx] = results.loading

        self.transformer_tap_module[tr_idx] = results.transformer_tap_module

        self.losses[br_idx] = results.losses

        self.flow_direction[br_idx] = results.flow_direction

        self.convergence_reports += results.convergence_reports

    def get_report_dataframe(self, island_idx=0):
        """
        Get a DataFrame containing the convergence report.

        Arguments:

            **island_idx**: (optional) island index

        Returns:

            DataFrame
        """
        report = self.convergence_reports[island_idx]
        data = {'Method': report.methods_,
                'Converged?': report.converged_,
                'Error': report.error_,
                'Elapsed (s)': report.elapsed_,
                'Iterations': report.iterations_}

        df = pd.DataFrame(data)

        return df

    def mdl(self, result_type: ResultTypes) -> "ResultsModel":
        """

        :param result_type:
        :return:
        """

        if result_type == ResultTypes.BusVoltageModule:
            labels = self.bus_names
            y = np.abs(self.voltage)
            y_label = '(p.u.)'
            title = 'Bus voltage '

        elif result_type == ResultTypes.BusVoltageAngle:
            labels = self.bus_names
            y = np.angle(self.voltage, deg=True)
            y_label = '(deg)'
            title = 'Bus voltage '

        elif result_type == ResultTypes.BusActivePower:
            labels = self.bus_names
            y = self.Sbus.real
            y_label = '(MW)'
            title = 'Bus Active power '

        elif result_type == ResultTypes.BusReactivePower:
            labels = self.bus_names
            y = self.Sbus.imag
            y_label = '(MVAr)'
            title = 'Bus Reactive power '

        elif result_type == ResultTypes.BusVoltagePolar:
            labels = self.bus_names
            y = self.voltage
            y_label = '(p.u.)'
            title = 'Bus voltage '

        elif result_type == ResultTypes.BranchPower:
            labels = self.branch_names
            y = self.Sf
            y_label = '(MVA)'
            title = 'Branch power '

        elif result_type == ResultTypes.BranchActivePowerFrom:
            labels = self.branch_names
            y = self.Sf.real
            y_label = '(MW)'
            title = 'Branch active power '

        elif result_type == ResultTypes.BranchReactivePowerFrom:
            labels = self.branch_names
            y = self.Sf.imag
            y_label = '(MVAr)'
            title = 'Branch reactive power '

        elif result_type == ResultTypes.BranchActivePowerTo:
            labels = self.branch_names
            y = self.St.real
            y_label = '(MW)'
            title = 'Branch active power '

        elif result_type == ResultTypes.BranchReactivePowerTo:
            labels = self.branch_names
            y = self.St.imag
            y_label = '(MVAr)'
            title = 'Branch reactive power '

        elif result_type == ResultTypes.Transformer2WTapModule:
            labels = self.transformer_names
            y = self.transformer_tap_module
            y_label = '(p.u.)'
            title = 'Transformer tap module '

        elif result_type == ResultTypes.BranchCurrent:
            labels = self.branch_names
            y = self.If
            y_label = '(p.u.)'
            title = 'Branch current '

        elif result_type == ResultTypes.BranchActiveCurrentFrom:
            labels = self.branch_names
            y = self.If.real
            y_label = '(p.u.)'
            title = 'Branch active current '

        elif result_type == ResultTypes.BranchReactiveCurrentFrom:
            labels = self.branch_names
            y = self.If.imag
            y_label = '(p.u.)'
            title = 'Branch reactive current '

        elif result_type == ResultTypes.BranchActiveCurrentTo:
            labels = self.branch_names
            y = self.It.real
            y_label = '(p.u.)'
            title = 'Branch active current '

        elif result_type == ResultTypes.BranchReactiveCurrentTo:
            labels = self.branch_names
            y = self.It.imag
            y_label = '(p.u.)'
            title = 'Branch reactive current '

        elif result_type == ResultTypes.BranchLoading:
            labels = self.branch_names
            y = np.abs(self.loading) * 100
            y_label = '(%)'
            title = 'Branch loading '

        elif result_type == ResultTypes.BranchLosses:
            labels = self.branch_names
            y = self.losses
            y_label = '(MVA)'
            title = 'Branch losses '

        elif result_type == ResultTypes.BranchActiveLosses:
            labels = self.branch_names
            y = self.losses.real
            y_label = '(MW)'
            title = 'Branch active losses '

        elif result_type == ResultTypes.BranchReactiveLosses:
            labels = self.branch_names
            y = self.losses.imag
            y_label = '(MVAr)'
            title = 'Branch reactive losses '

        elif result_type == ResultTypes.BranchVoltage:
            labels = self.branch_names
            y = np.abs(self.Vbranch)
            y_label = '(p.u.)'
            title = 'Branch voltage drop '

        elif result_type == ResultTypes.BranchAngles:
            labels = self.branch_names
            y = np.angle(self.Vbranch, deg=True)
            y_label = '(deg)'
            title = 'Branch voltage angle '

        elif result_type == ResultTypes.BranchTapModule:
            labels = self.branch_names
            y = self.ma
            y_label = '(p.u.)'
            title = 'Branch tap module '

        elif result_type == ResultTypes.BranchTapAngle:
            labels = self.branch_names
            y = self.theta
            y_label = '(rad)'
            title = 'Branch tap angle '

        elif result_type == ResultTypes.BranchBeq:
            labels = self.branch_names
            y = self.Beq
            y_label = '(p.u.)'
            title = 'Branch Beq '

        elif result_type == ResultTypes.HvdcLosses:
            labels = self.hvdc_names
            y = self.hvdc_losses
            y_label = '(MW)'
            title = result_type.value

        elif result_type == ResultTypes.HvdcPowerFrom:
            labels = self.hvdc_names
            y = self.hvdc_Pf
            y_label = '(MW)'
            title = result_type.value

        elif result_type == ResultTypes.HvdcPowerTo:
            labels = self.hvdc_names
            y = self.hvdc_Pt
            y_label = '(MW)'
            title = result_type.value

        else:
            raise Exception('Unsupported result type: ' + str(result_type))

        # assemble model
        mdl = ResultsModel(data=y, index=labels, columns=[result_type.value[0]],
                           title=title, ylabel=y_label, units=y_label)
        return mdl

    def export_all(self):
        """
        Exports all the results to DataFrames.

        Returns:

            Bus results, Branch reuslts
        """

        # buses results
        vm = np.abs(self.voltage)
        va = np.angle(self.voltage)
        vr = self.voltage.real
        vi = self.voltage.imag
        bus_data = np.c_[vr, vi, vm, va]
        bus_cols = ['Real voltage (p.u.)', 'Imag Voltage (p.u.)', 'Voltage module (p.u.)', 'Voltage angle (rad)']
        df_bus = pd.DataFrame(data=bus_data, columns=bus_cols)

        # branch results
        sr = self.Sf.real
        si = self.Sf.imag
        sm = np.abs(self.Sf)
        ld = np.abs(self.loading)
        la = self.losses.real
        lr = self.losses.imag
        ls = np.abs(self.losses)
        tm = self.transformer_tap_module

        branch_data = np.c_[sr, si, sm, ld, la, lr, ls, tm]
        branch_cols = ['Real power (MW)', 'Imag power (MVAr)', 'Power module (MVA)', 'Loading(%)',
                       'Losses (MW)', 'Losses (MVAr)', 'Losses (MVA)', 'Tap module']
        df_branch = pd.DataFrame(data=branch_data, columns=branch_cols)

        return df_bus, df_branch




# This file is part of GridCal.
#
# GridCal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GridCal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GridCal.  If not, see <http://www.gnu.org/licenses/>.
import numpy as np


class BusData:

    def __init__(self, nbus, ntime=1):
        """

        :param nbus:
        """
        self.nbus = nbus
        self.ntime = ntime
        self.bus_names = np.empty(nbus, dtype=object)
        self.bus_active = np.ones((nbus, ntime), dtype=int)
        self.Vbus = np.ones((nbus, ntime), dtype=complex)
        self.Vmin = np.ones(nbus, dtype=float)
        self.Vmax = np.ones(nbus, dtype=float)
        self.bus_types = np.empty(nbus, dtype=int)
        self.bus_installed_power = np.zeros(nbus, dtype=float)
        self.bus_is_dc = np.empty(nbus, dtype=bool)

    def slice(self, elm_idx, time_idx=None):
        """
        Slice this data structure
        :param elm_idx: array of bus indices
        :param time_idx: array of time indices
        :return: instance of BusData
        """

        if time_idx is None:
            tidx = elm_idx
        else:
            tidx = np.ix_(elm_idx, time_idx)

        data = BusData(nbus=len(elm_idx))

        data.bus_names = self.bus_names[elm_idx]

        data.bus_active = self.bus_active[tidx]
        data.Vbus = self.Vbus[tidx]
        data.Vmin = self.Vmin[elm_idx]
        data.Vmax = self.Vmax[elm_idx]

        data.bus_types = self.bus_types[elm_idx]
        data.bus_installed_power = self.bus_installed_power[elm_idx]
        data.bus_is_dc = self.bus_is_dc[elm_idx]

        return data

    def __len__(self):
        return self.nbus




"""
.. module:: tcpserve
   :synopsis: TCP-based controller server and workers for POAP.
.. moduleauthor:: David Bindel <bindel@cornell.edu>
"""

import time
import socket
import threading
import pickle
import logging

try:
    import socketserver
except ImportError:
    import SocketServer as socketserver

from poap.controller import ThreadController


logger = logging.getLogger(__name__)


class SocketWorkerHandler(socketserver.BaseRequestHandler):
    """Manage a remote worker for a thread controller.

    The SocketWorkerHandler is a request handler for incoming workers.
    It implements the socketserver request handler interface, and also
    the worker interface expected by the ThreadController.
    """

    def eval(self, record):
        "Send an evaluation request to remote worker"
        logger.debug("Send eval to worker")
        self.records[id(record)] = record
        try:
            if record.extra_args is None:
                m = ('eval', id(record), record.params)
            else:
                m = ('eval', id(record), record.params, record.extra_args)
            self.request.send(self.server.marshall(*m))
        except Exception as e:
            logger.warning("In eval: {0}".format(e))
            self._cleanup(record)

    def kill(self, record):
        "Send a kill request to a remote worker"
        logger.debug("Send kill to worker")
        try:
            self.request.send(
                self.server.marshall('kill', id(record)))
        except socket.error as e:
            logger.warning("In kill: {0}".format(e))
            self._cleanup(record)

    def terminate(self):
        "Send a termination request to a remote worker"
        if not self.running:
            return
        logger.debug("Send terminate to worker")
        try:
            self.running = False
            self.request.send(self.server.marshall('terminate'))
            self.request.close()
        except socket.error as e:
            logger.warning("In terminate: {0}".format(e))

    def _handle_message(self, args):
        "Receive a record status message"
        mname = args[0]
        record = self.records[args[1]]
        controller = self.server.controller
        if mname in self.server.message_handlers:
            handler = self.server.message_handlers[mname]
            controller.add_message(lambda: handler(record, *args[2:]))
        else:
            method = getattr(record, mname)
            controller.add_message(lambda: method(*args[2:]))
        if mname == 'complete' or mname == 'cancel' or mname == 'kill':
            logger.debug("Re-queueing worker")
            controller.add_worker(self)

    def _cleanup(self, record):
        "Clean up an incomplete record assigned to this worker."
        def killrec():
            if not record.is_done:
                logger.debug("Kill {0}".format(record.params))
                record.kill()
        self.server.controller.add_message(killrec)

    def handle(self):
        "Main event loop called from SocketServer"
        self.records = {}
        self.running = True
        self.server.controller.add_term_callback(self.terminate)
        try:
            self.server.controller.add_worker(self)
            while self.running:
                logger.debug("Waiting for worker input")
                data = self.request.recv(4096)
                if not data:
                    return
                args = self.server.unmarshall(data)
                self._handle_message(args)
        except socket.error as e:
            logger.debug("Exiting worker: {0}".format(e))
        finally:
            for rec_id, record in self.records.items():
                self._cleanup(record)
            logger.debug("Leaving worker thread")


class ThreadedTCPServer(socketserver.ThreadingMixIn,
                        socketserver.TCPServer, object):
    """SocketServer interface for workers to connect to controller.

    The socket server interface lets workers connect to a given
    TCP/IP port and exchange updates with the controller.

    The server sends messages of the form

        ('eval', record_id, args, extra_args)
        ('eval', record_id, args)
        ('kill', record_id)
        ('terminate')

    The default messages received are

        ('running', record_id)
        ('kill', record_id)
        ('cancel', record_id)
        ('complete', record_id, value)

    The set of handlers can also be extended with a dictionary of
    named callbacks to be invoked whenever a record update comes in.
    For example, to set a lower bound field, we might use the handler

        def set_lb(rec, value):
            rec.lb = value
        handlers = {'lb' : set_lb }

    This is useful for adding new types of updates without mucking
    around in the EvalRecord implementation.

    Attributes:
        controller: ThreadController that manages the optimization
        handlers: dictionary of specialized message handlers
        strategy: redirects to the controller strategy
    """

    def __init__(self, sockname=("localhost", 0), strategy=None, handlers={}):
        """Initialize the controller on the given (host,port) address

        Args:
            sockname: Socket on which to serve workers
            strategy: Strategy object to connect to controllers
            handlers: Dictionary of specialized message handlers
        """
        super(ThreadedTCPServer, self).__init__(sockname, SocketWorkerHandler)
        self.message_handlers = handlers
        self.controller = ThreadController()
        self.controller.strategy = strategy
        self.controller.add_term_callback(self.shutdown)

    def marshall(self, *args):
        "Convert an argument list to wire format."
        return pickle.dumps(args)

    def unmarshall(self, data):
        "Convert wire format back to Python arg list."
        return pickle.loads(data)

    @property
    def strategy(self):
        return self.controller.strategy

    @strategy.setter
    def strategy(self, strategy):
        self.controller.strategy = strategy

    @property
    def sockname(self):
        return self.socket.getsockname()

    def run(self, merit=lambda r: r.value, filter=None):
        thread = threading.Thread(target=self.controller.run)
        thread.start()
        self.serve_forever()
        thread.join()
        return self.controller.best_point(merit=merit, filter=filter)


class SocketWorker(object):
    """Base class for workers that connect to SocketServer interface

    The socket server interface is a server to which workers can
    connect.  The socket worker is the client to that interface.  It
    connects to a given TCP/IP port, then attempts to do work on the
    controller's behalf.

    Attributes:
        running: True if the socket is active
        sock: Worker TCP socket
    """

    def __init__(self, sockname, retries=0):
        """Initialize the SocketWorker.

        The constructor tries to open the socket; on failure, it keeps
        trying up to retries times, once per second.

        Args:
            sockname: (host, port) tuple where server lives
            retries: number of times to retry the connection
        """
        self.running = False
        while not self.running and retries >= 0:
            try:
                self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                self.sock.connect(sockname)
                self.running = True
            except socket.error as e:
                logger.warning("Worker could not connect: {0}".format(e))
                retries -= 1
                time.sleep(1)

    def marshall(self, *args):
        "Marshall data to wire format"
        return pickle.dumps(args)

    def unmarshall(self, data):
        "Convert data from wire format back to Python tuple"
        return pickle.loads(data)

    def send(self, *args):
        "Send a message to the controller"
        self.sock.send(self.marshall(*args))

    def _run(self):
        "Run a message from the controller"
        if not self.running:
            return
        data = self.unmarshall(self.sock.recv(4096))
        method = getattr(self, data[0])
        method(*data[1:])

    def eval(self, record_id, params):
        "Compute a function value"
        pass

    def kill(self, record_id):
        "Kill a function evaluation"
        pass

    def terminate(self):
        "Terminate the worker"
        self.running = False

    def run(self, loop=True):
        "Main loop"
        try:
            self._run()
            while loop and self.running:
                self._run()
        except socket.error as e:
            logger.warning("Exit loop: {0}".format(e))
        finally:
            self.sock.close()


class SimpleSocketWorker(SocketWorker):
    """Simple socket worker that runs a local objective function

    The SimpleSocketWorker is a socket worker that runs a local Python
    function and returns the result.  It is probably mostly useful for
    testing -- the ProcessSocketWorker is a better option for external
    simulations.
    """

    def __init__(self, objective, sockname, retries=0):
        """Initialize the SimpleSocketWorker.

        The constructor tries to open the socket; on failure, it keeps
        trying up to retries times, once per second.

        Args:
            objective: Python objective function
            sockname: (host, port) tuple where server lives
            retries: number of times to retry the connection
        """
        SocketWorker.__init__(self, sockname, retries)
        self.objective = objective

    def eval(self, record_id, params):
        """Evaluate the function and send back a result.

        If the function evaluation crashes, we send back a 'cancel'
        request for the record.  If, on the other hand, there is a
        problem with calling send, we probably want to let the worker
        error out.

        Args:
            record_id: Feval record identifier used by server/controller
            params: Parameters sent to the function to be evaluated
        """
        try:
            msg = ('complete', record_id, self.objective(*params))
        except:
            msg = ('cancel', record_id)
        self.send(*msg)


class ProcessSocketWorker(SocketWorker):
    """Socket worker that runs an evaluation in a subprocess

    The ProcessSocketWorker is a base class for simulations that run a
    simulation in an external subprocess.  This class provides functionality
    just to allow graceful termination of the external simulations.

    Attributes:
        process: Handle for external subprocess
    """

    def __init__(self, sockname, retries=0):
        """Initialize the ProcessSocketWorker.

        The constructor tries to open the socket; on failure, it keeps
        trying up to retries times, once per second.

        Args:
            sockname: (host, port) tuple where server lives
            retries: number of times to retry the connection
        """
        SocketWorker.__init__(self, sockname, retries)
        self.process = None

    def kill_process(self):
        "Kill the child process"
        if self.process is not None and self.process.poll() is None:
            logger.debug("ProcessSocketWorker is killing subprocess")
            self.process.terminate()

    def kill(self, record_id):
        self.kill_process()

    def terminate(self):
        self.kill_process()
        SocketWorker.terminate(self)



# This file is part of GridCal.
#
# GridCal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GridCal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GridCal.  If not, see <http://www.gnu.org/licenses/>.
import numpy as np
from PySide2.QtWidgets import *
from PySide2.QtCore import *
from PySide2.QtGui import *
from GridCal.Engine.Devices.generator import Generator
from GridCal.Gui.GridEditorWidget.generic_graphics import ACTIVE, DEACTIVATED, OTHER, Circle
from GridCal.Gui.GuiFunctions import ObjectsModel, PandasModel
from GridCal.Gui.GridEditorWidget.messages import *
from GridCal.Gui.GridEditorWidget.matplotlibwidget import MatplotlibWidget


class GeneratorEditor(QDialog):

    def __init__(self, generator: Generator):
        """
        Line Editor constructor
        :param generator: Generator object to update
        """
        super(GeneratorEditor, self).__init__()

        # keep pointer to the line object
        self.generator = generator

        self.selected_template = None

        self.setObjectName("self")

        self.setContextMenuPolicy(Qt.NoContextMenu)

        self.layout = QVBoxLayout(self)

        # create matplotlib object
        self.plotter = MatplotlibWidget(parent=self)
        self.layout.addWidget(self.plotter)

        self.setLayout(self.layout)

        self.plot_q_points()

    def plot_q_points(self):
        p = self.generator.q_points[:, 0]
        qmin = self.generator.q_points[:, 1]
        qmax = self.generator.q_points[:, 2]
        self.plotter.plot(qmax, p, 'x-')
        self.plotter.plot(qmin, p, 'x-')
        self.plotter.redraw()


class GeneratorGraphicItem(QGraphicsItemGroup):

    def __init__(self, parent, api_obj, diagramScene):
        """

        :param parent:
        :param api_obj:
        """
        super(GeneratorGraphicItem, self).__init__(parent)

        self.parent = parent

        self.api_object = api_obj

        self.diagramScene = diagramScene

        self.w = 40
        self.h = 40

        # Properties of the container:
        self.setFlags(self.ItemIsSelectable | self.ItemIsMovable)
        self.setCursor(QCursor(Qt.PointingHandCursor))

        self.width = 4
        if self.api_object is not None:
            if self.api_object.active:
                self.style = ACTIVE['style']
                self.color = ACTIVE['color']
            else:
                self.style = DEACTIVATED['style']
                self.color = DEACTIVATED['color']
        else:
            self.style = OTHER['style']
            self.color = OTHER['color']

        # line to tie this object with the original bus (the parent)
        self.nexus = QGraphicsLineItem()
        self.nexus.setPen(QPen(self.color, self.width, self.style))
        parent.scene().addItem(self.nexus)

        pen = QPen(self.color, self.width, self.style)

        self.glyph = Circle(self)
        self.glyph.setRect(0, 0, self.h, self.w)
        self.glyph.setPen(pen)
        self.addToGroup(self.glyph)

        self.label = QGraphicsTextItem('G', parent=self.glyph)
        self.label.setDefaultTextColor(self.color)
        self.label.setPos(self.h / 4, self.w / 5)

        self.setPos(self.parent.x(), self.parent.y() + 100)
        self.update_line(self.pos())

    def update_line(self, pos):
        """
        Update the line that joins the parent and this object
        :param pos: position of this object
        """
        parent = self.parentItem()
        rect = parent.rect()
        self.nexus.setLine(
            pos.x() + self.w / 2,
            pos.y() + 0,
            parent.x() + rect.width() / 2,
            parent.y() + parent.terminal.y() + 5,
        )
        self.setZValue(-1)
        self.nexus.setZValue(-1)

    def contextMenuEvent(self, event):
        """
        Display context menu
        @param event:
        @return:
        """
        menu = QMenu()
        menu.addSection("Generator")

        pe = menu.addAction('Active')
        pe.setCheckable(True)
        pe.setChecked(self.api_object.active)
        pe.triggered.connect(self.enable_disable_toggle)

        pc = menu.addAction('Voltage control')
        pc.setCheckable(True)
        pc.setChecked(self.api_object.is_controlled)
        pc.triggered.connect(self.enable_disable_control_toggle)

        pa = menu.addAction('Plot profiles')
        plot_icon = QIcon()
        plot_icon.addPixmap(QPixmap(":/Icons/icons/plot.svg"))
        pa.setIcon(plot_icon)
        pa.triggered.connect(self.plot)

        da = menu.addAction('Delete')
        del_icon = QIcon()
        del_icon.addPixmap(QPixmap(":/Icons/icons/delete3.svg"))
        da.setIcon(del_icon)
        da.triggered.connect(self.remove)

        menu.exec_(event.screenPos())

    def remove(self):
        """
        Remove this element
        @return:
        """
        ok = yes_no_question('Are you sure that you want to remove this generator', 'Remove generator')
        if ok:
            self.diagramScene.removeItem(self.nexus)
            self.diagramScene.removeItem(self)
            self.api_object.bus.controlled_generators.remove(self.api_object)

    def enable_disable_toggle(self):
        """

        @return:
        """
        if self.api_object is not None:
            if self.api_object.active:
                self.set_enable(False)
            else:
                self.set_enable(True)

    def enable_disable_control_toggle(self):
        """
        Enable / Disable device voltage control
        """
        if self.api_object is not None:
            self.api_object.is_controlled = not self.api_object.is_controlled

    def set_enable(self, val=True):
        """
        Set the enable value, graphically and in the API
        @param val:
        @return:
        """
        self.api_object.active = val
        if self.api_object is not None:
            if self.api_object.active:
                self.style = ACTIVE['style']
                self.color = ACTIVE['color']
            else:
                self.style = DEACTIVATED['style']
                self.color = DEACTIVATED['color']
        else:
            self.style = OTHER['style']
            self.color = OTHER['color']
        self.glyph.setPen(QPen(self.color, self.width, self.style))
        self.label.setDefaultTextColor(self.color)

    def plot(self):
        """
        Plot API objects profiles
        """
        # time series object from the last simulation
        ts = self.diagramScene.circuit.time_profile

        # plot the profiles
        self.api_object.plot_profiles(time=ts)

    def edit(self):
        """
        Open the appropriate editor dialogue
        :return:
        """
        dlg = GeneratorEditor(self.api_object)
        if dlg.exec_():
            pass

    def mousePressEvent(self, QGraphicsSceneMouseEvent):
        """
        mouse press: display the editor
        :param QGraphicsSceneMouseEvent:
        :return:
        """
        mdl = ObjectsModel([self.api_object], self.api_object.editable_headers,
                           parent=self.diagramScene.parent().object_editor_table, editable=True, transposed=True)
        self.diagramScene.parent().object_editor_table.setModel(mdl)

    def mouseDoubleClickEvent(self, event):
        self.edit()




from enum import Enum
import numpy as np


class BusTypes(Enum):
    Ref = 0,
    PQ = 1,
    PV = 2


class Connection(Enum):
    Delta = 0,  # Delta connection
    Wye = 1,  # Wye (Star) connection isolated from ground
    WyeG = 2  # Wye (Star) connection connected from ground
    PositiveSequence = 3  # When running a positive sequence the connection does not matter


class LoadTypes(Enum):
    ZIP = 0,
    Exponential = 1,
    Polynomial = 2


class Phases:
    ABC = [0, 1, 2]
    ACB = [0, 2, 1]
    CBA = [2, 1, 0]
    A = [0]
    B = [1]
    C = [2]
    AB = [0, 1]
    BA = [1, 0]
    AC = [0, 2]
    CA = [2, 0]
    BC = [1, 2]
    CB = [2, 1]


class PowerFlowMethods(Enum):
    GaussSeidel = 1,
    NewtonRaphson = 2,
    LevenbergMarquardt = 3,
    LinearAC = 4,
    GaussRaphson = 5,
    ZMatrix = 6


def seq_to_abc(seq):
    """
    Convert sequence vector to ABC full matrix
    Kersting p.136 / Computer Analysis of power systems p. 43
    :param seq: vector of the 0, 1, 2 sequence components
    :return: 3x3 matrix
    """

    if len(seq) == 2:
        k1 = (2.0 * seq[1] + seq[0]) / 3.0
        k2 = (seq[0] - seq[1]) / 3.0
        return np.array([[k1, k2, k2], [k2, k1, k2], [k2, k2, k1]])
    elif len(seq) == 3:
        a = np.exp(2j * np.pi / 3)
        a2 = a * a
        k1 = seq[0] + seq[1] + seq[2]
        k2 = seq[0] + a * seq[1] + a2 * seq[2]
        k3 = seq[0] + a2 * seq[1] + a * seq[2]
        return np.array([[k1, k2, k3], [k3, k1, k2], [k2, k3, k1]])


def wye_to_delta(vector):
    """
    Transform Wye connection to Delta connection
    :param vector: vector of 3 values
    :return: Delta connected values
    """
    D = np.array([[1, -1, 0], [0, 1, -1], [-1, 0, 1]])
    return D.dot(vector)


def delta_to_wye(vector):
    """
    Transform Delta connection to Wye connection
    :param vector: vector of 3 values
    :return: Wye connected values
    """
    D = (1 / 3) * np.array([[1, 0, -1], [-1, 1, 0], [0, -1, 1]])
    return D.dot(vector)




# This file is part of GridCal.
#
# GridCal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GridCal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GridCal.  If not, see <http://www.gnu.org/licenses/>.
from enum import Enum
import numpy as np
import time
from PySide2.QtCore import QThread, Signal

from GridCal.Engine.basic_structures import Logger
from GridCal.Engine.basic_structures import TimeGrouping, MIPSolvers
from GridCal.Engine.Simulations.OPF.opf_results import OptimalPowerFlowResults
from GridCal.Engine.Core.multi_circuit import MultiCircuit
from GridCal.Engine.Simulations.OPF.ac_opf import OpfAc
from GridCal.Engine.Simulations.OPF.dc_opf import OpfDc
from GridCal.Engine.Simulations.OPF.simple_dispatch import OpfSimple
from GridCal.Engine.basic_structures import SolverType
from GridCal.Engine.Simulations.PowerFlow.power_flow_driver import PowerFlowOptions
from GridCal.Engine.Core.snapshot_opf_data import compile_snapshot_opf_circuit

########################################################################################################################
# Optimal Power flow classes
########################################################################################################################


class OptimalPowerFlowOptions:

    def __init__(self, verbose=False,
                 solver: SolverType = SolverType.DC_OPF,
                 grouping: TimeGrouping = TimeGrouping.NoGrouping,
                 mip_solver=MIPSolvers.CBC,
                 faster_less_accurate=False,
                 power_flow_options=None, bus_types=None):
        """
        Optimal power flow options
        :param verbose:
        :param solver: OPF solver type
        :param grouping:
        :param mip_solver:
        :param faster_less_accurate:
        :param power_flow_options:
        :param bus_types:
        """
        self.verbose = verbose

        self.solver = solver

        self.grouping = grouping

        self.mip_solver = mip_solver

        self.faster_less_accurate = faster_less_accurate

        self.power_flow_options = power_flow_options

        self.bus_types = bus_types


class OptimalPowerFlow(QThread):
    progress_signal = Signal(float)
    progress_text = Signal(str)
    done_signal = Signal()
    name = 'Optimal power flow'

    def __init__(self, grid: MultiCircuit, options: OptimalPowerFlowOptions, pf_options: PowerFlowOptions):
        """
        PowerFlowDriver class constructor
        @param grid: MultiCircuit Object
        @param options: OPF options
        """
        QThread.__init__(self)

        # Grid to run a power flow in
        self.grid = grid

        # Options to use
        self.options = options

        self.pf_options = pf_options

        # OPF results
        self.results = None

        # set cancel state
        self.__cancel__ = False

        self.all_solved = True

        self.elapsed = 0.0

        self.logger = Logger()

    def get_steps(self):
        """
        Get time steps list of strings
        """
        return list()

    def opf(self):
        """
        Run a power flow for every circuit
        @return: OptimalPowerFlowResults object
        """

        numerical_circuit = compile_snapshot_opf_circuit(circuit=self.grid,
                                                         apply_temperature=self.pf_options.apply_temperature_correction,
                                                         branch_tolerance_mode=self.pf_options.branch_impedance_tolerance_mode)

        if self.options.solver == SolverType.DC_OPF:
            # DC optimal power flow
            problem = OpfDc(numerical_circuit=numerical_circuit, solver=self.options.mip_solver)

        elif self.options.solver == SolverType.AC_OPF:
            # AC optimal power flow
            problem = OpfAc(numerical_circuit=numerical_circuit, solver=self.options.mip_solver)

        elif self.options.solver == SolverType.Simple_OPF:
            # simplistic dispatch
            problem = OpfSimple(numerical_circuit=numerical_circuit)

        else:
            raise Exception('Solver not recognized ' + str(self.options.solver))

        # Solve
        problem.solve()

        # get the branch flows (it is used more than one time)
        Sbr = problem.get_branch_power()
        ld = problem.get_load_shedding()
        ld[ld == None] = 0
        bt = problem.get_battery_power()
        bt[bt == None] = 0
        gn = problem.get_generator_power()
        gn[gn == None] = 0

        # pack the results
        self.results = OptimalPowerFlowResults(bus_names=numerical_circuit.bus_data.bus_names,
                                               branch_names=numerical_circuit.branch_data.branch_names,
                                               load_names=numerical_circuit.load_data.load_names,
                                               generator_names=numerical_circuit.generator_data.generator_names,
                                               battery_names=numerical_circuit.battery_data.battery_names,
                                               Sbus=None,
                                               voltage=problem.get_voltage(),
                                               load_shedding=ld,
                                               generation_shedding=np.zeros_like(gn),
                                               battery_power=bt,
                                               controlled_generation_power=gn,
                                               Sf=Sbr,
                                               overloads=problem.get_overloads(),
                                               loading=problem.get_loading(),
                                               converged=bool(problem.converged()),
                                               bus_types=numerical_circuit.bus_types)

        return self.results

    def run(self):
        """

        :return:
        """
        start = time.time()
        self.opf()
        end = time.time()
        self.elapsed = end - start
        self.progress_text.emit('Done!')
        self.done_signal.emit()

    def cancel(self):
        self.__cancel__ = True





from research.three_phase.Engine.general import *
from research.three_phase.Engine.Devices.generator import *
from research.three_phase.Engine.Devices.load import *
from research.three_phase.Engine.Devices.line import *
from research.three_phase.Engine.Devices.transformer import *
from research.three_phase.Engine.Devices.bus import *
from research.three_phase.Engine.circuit import *
from research.three_phase.Engine.Simulation.power_flow import *



# Copyright (c) 1996-2015 PSERC. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE_MATPOWER file.

# The file has been modified from Pypower.
# The function mu() has been added to the solver in order to provide an optimal iteration control
# Copyright (c) 2016 Santiago Peñate Vera
# This file retains the BSD-Style license

import numpy as np
from numpy import array, angle, exp, linalg, r_, Inf, conj, diag, asmatrix, asarray, zeros_like
from scipy.sparse import issparse, csr_matrix as sparse, hstack, vstack
from scipy.sparse.linalg import spsolve
import scipy
scipy.ALLOW_THREADS = True

np.set_printoptions(precision=8, suppress=True, linewidth=320)


def dSbus_dV(Ybus, V, I):
    """
    Computes partial derivatives of power injection w.r.t. voltage.
    :param Ybus: Admittance matrix
    :param V: Bus voltages array
    :param I: Bus current injections array
    :return:
    """
    '''
    Computes partial derivatives of power injection w.r.t. voltage.

    Returns two matrices containing partial derivatives of the complex bus
    power injections w.r.t voltage magnitude and voltage angle respectively
    (for all buses). If C{Ybus} is a sparse matrix, the return values will be
    also. The following explains the expressions used to form the matrices::

        Ibus = Ybus * V - I

        S = diag(V) * conj(Ibus) = diag(conj(Ibus)) * V

    Partials of V & Ibus w.r.t. voltage magnitudes::
        dV/dVm = diag(V / abs(V))
        dI/dVm = Ybus * dV/dVm = Ybus * diag(V / abs(V))

    Partials of V & Ibus w.r.t. voltage angles::
        dV/dVa = j * diag(V)
        dI/dVa = Ybus * dV/dVa = Ybus * j * diag(V)

    Partials of S w.r.t. voltage magnitudes::
        dS/dVm = diag(V) * conj(dI/dVm) + diag(conj(Ibus)) * dV/dVm
               = diag(V) * conj(Ybus * diag(V / abs(V)))
                                        + conj(diag(Ibus)) * diag(V / abs(V))

    Partials of S w.r.t. voltage angles::
        dS/dVa = diag(V) * conj(dI/dVa) + diag(conj(Ibus)) * dV/dVa
               = diag(V) * conj(Ybus * j * diag(V))
                                        + conj(diag(Ibus)) * j * diag(V)
               = -j * diag(V) * conj(Ybus * diag(V))
                                        + conj(diag(Ibus)) * j * diag(V)
               = j * diag(V) * conj(diag(Ibus) - Ybus * diag(V))

    For more details on the derivations behind the derivative code used
    in PYPOWER information, see:

    [TN2]  R. D. Zimmerman, "AC Power Flows, Generalized OPF Costs and
    their Derivatives using Complex Matrix Notation", MATPOWER
    Technical Note 2, February 2010.
    U{http://www.pserc.cornell.edu/matpower/TN2-OPF-Derivatives.pdf}

    @author: Ray Zimmerman (PSERC Cornell)
    '''

    ib = range(len(V))

    if issparse(Ybus):
        Ibus = Ybus * V - I

        diagV = sparse((V, (ib, ib)))
        diagIbus = sparse((Ibus, (ib, ib)))
        diagVnorm = sparse((V / abs(V), (ib, ib)))
    else:
        Ibus = Ybus * asmatrix(V).T - I

        diagV = asmatrix(diag(V))
        diagIbus = asmatrix(diag(asarray(Ibus).flatten()))
        diagVnorm = asmatrix(diag(V / abs(V)))

    dS_dVm = diagV * conj(Ybus * diagVnorm) + conj(diagIbus) * diagVnorm
    dS_dVa = 1j * diagV * conj(diagIbus - Ybus * diagV)

    return dS_dVm, dS_dVa


def Jacobian(Ybus, V, Ibus, pq, pvpq):
    """
    Computes the system Jacobian matrix
    Args:
        Ybus: Admittance matrix
        V: Array of nodal voltages
        Ibus: Array of nodal current injections
        pq: Array with the indices of the PQ buses
        pvpq: Array with the indices of the PV and PQ buses

    Returns:
        The system Jacobian matrix
    """
    dS_dVm, dS_dVa = dSbus_dV(Ybus, V, Ibus)  # compute the derivatives

    J11 = dS_dVa[array([pvpq]).T, pvpq].real
    J12 = dS_dVm[array([pvpq]).T, pq].real
    J21 = dS_dVa[array([pq]).T, pvpq].imag
    J22 = dS_dVm[array([pq]).T, pq].imag

    J = vstack([
        hstack([J11, J12]),
        hstack([J21, J22])
    ], format="csr")

    return J


def LevenbergMarquardtPF(Ybus, Sbus, V0, Ibus, pv, pq, tol, max_it):
    """
    Solves the power flow problem by the Levenberg-Marquardt power flow algorithm
    Args:
        Ybus: Admittance matrix
        Sbus: Array of nodal power injections
        V0: Array of nodal voltages (initial solution)
        Ibus: Array of nodal current injections
        ref: Array with the indices of the slack buses
        pv: Array with the indices of the PV buses
        pq: Array with the indices of the PQ buses
        tol: Tolerance
        max_it: Maximum number of iterations
        verbose: Boolean variable for the verbose mode activation
    Returns:
    @Author: Santiago Penate Vera
    """

    # initialize
    V = V0
    Va = angle(V)
    Vm = abs(V)
    dVa = zeros_like(Va)
    dVm = zeros_like(Vm)
    # set up indexing for updating V
    pvpq = r_[pv, pq]
    npv = len(pv)
    npq = len(pq)

    # j1:j2 - V angle of pv buses
    j1 = 0
    j2 = npv
    # j3:j4 - V angle of pq buses
    j3 = j2
    j4 = j2 + npq
    # j5:j6 - V mag of pq buses
    j5 = j4
    j6 = j4 + npq

    update_H = True
    converged = False
    iter_ = 0
    nu = 2.0
    lbmda = 0
    f_prev = 1e9  # very large number
    Idn = np.identity(2 * npq + npv)

    # do Newton iterations
    while not converged and iter_ < max_it:

        # evaluate Jacobian
        if update_H:
            H = Jacobian(Ybus, V, Ibus, pq, pvpq)

        # evaluate the solution error F(x0)
        Scalc = V * conj(Ybus * V - Ibus)
        mis = Scalc - Sbus  # compute the mismatch
        dz = r_[mis[pv].real, mis[pq].real, mis[pq].imag]  # mismatch in the Jacobian order

        # system matrix
        # H1 = H^t·W
        H1 = H.transpose()
        # H2 = H1·H
        H2 = H1.dot(H)

        # set first value of lmbda
        if iter_ == 0:
            lbmda = 1e-3 * H2.diagonal().max()

        # compute system matrix
        A = H2 + lbmda * Idn

        # right hand side
        # H^t·W·dz
        rhs = H1.dot(dz)

        # Solve the increment
        dx = np.linalg.solve(A, rhs)

        # objective function
        f = 0.5 * dz.dot(dz)

        # decision function
        rho = (f_prev - f) / (0.5 * dx.dot(lbmda * dx + rhs))

        # lambda update
        if rho > 0:
            update_H = True
            lbmda *= max([1.0 / 3.0, 1 - (2 * rho - 1) ** 3])
            nu = 2.0

            # reassign the solution vector
            if npv:
                dVa[pv] = dx[j1:j2]
            if npq:
                dVa[pq] = dx[j3:j4]
                dVm[pq] = dx[j5:j6]

            Vm -= dVm
            Va -= dVa
            V = Vm * exp(1j * Va)
            Vm = abs(V)  # update Vm and Va again in case
            Va = angle(V)  # we wrapped around with a negative Vm
        else:
            update_H = False
            lbmda *= nu
            nu *= 2

        # ckeck convergence
        normF = np.linalg.norm(dx, np.Inf)
        converged = normF < tol
        f_prev = f

        # update iteration counter
        iter_ += 1

    return V, converged, normF, Scalc


if __name__ == "__main__":
    from GridCal.Engine.calculation_engine import *

    grid = MultiCircuit()
    # grid.load_file('lynn5buspq.xlsx')
    grid.load_file('IEEE30.xlsx')

    grid.compile()

    circuit = grid.circuits[0]

    print('\nYbus:\n', circuit.power_flow_input.Ybus.todense())
    print('\nYseries:\n', circuit.power_flow_input.Yseries.todense())
    print('\nYshunt:\n', circuit.power_flow_input.Yshunt)
    print('\nSbus:\n', circuit.power_flow_input.Sbus)
    print('\nIbus:\n', circuit.power_flow_input.Ibus)
    print('\nVbus:\n', circuit.power_flow_input.Vbus)
    print('\ntypes:\n', circuit.power_flow_input.types)
    print('\npq:\n', circuit.power_flow_input.pq)
    print('\npv:\n', circuit.power_flow_input.pv)
    print('\nvd:\n', circuit.power_flow_input.ref)

    import time
    print('Levenberg-Marquardt')
    start_time = time.time()
    V1, converged_, err, S = LevenbergMarquardtPF(Ybus=circuit.power_flow_input.Ybus,
                                                  Sbus=circuit.power_flow_input.Sbus,
                                                  V0=circuit.power_flow_input.Vbus,
                                                  Ibus=circuit.power_flow_input.Ibus,
                                                  pv=circuit.power_flow_input.pv,
                                                  pq=circuit.power_flow_input.pq,
                                                  tol=1e-9,
                                                  max_it=100)

    print("--- %s seconds ---" % (time.time() - start_time))
    # print_coeffs(C, W, R, X, H)

    print('V module:\t', abs(V1))
    print('V angle: \t', angle(V1))
    print('error: \t', err)

    # check the HELM solution: v against the NR power flow
    print('\nNR')
    options = PowerFlowOptions(SolverType.NR, verbose=False, robust=False, tolerance=1e-9)
    power_flow = PowerFlow(grid, options)

    start_time = time.time()
    power_flow.run()
    print("--- %s seconds ---" % (time.time() - start_time))
    vnr = circuit.power_flow_results.voltage

    print('V module:\t', abs(vnr))
    print('V angle: \t', angle(vnr))
    print('error: \t', circuit.power_flow_results.error)

    # check
    print('\ndiff:\t', V1 - vnr)


# -*- coding: utf-8 -*-
# This file is part of GridCal.
#
# GridCal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GridCal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GridCal.  If not, see <http://www.gnu.org/licenses/>.

import Desarrollos.power_flow_research.example_grids as grids
import numpy as np
np.set_printoptions(linewidth=320)
# np.set_printoptions(precision=6, suppress=True, linewidth=320)
from numpy import where, zeros, ones, mod, conj, array, dot, complex128
from numpy import poly1d, r_, eye, hstack, diag, linalg, Inf
from enum import Enum
from itertools import product

from scipy import fftpack
from scipy.linalg import solve

from scipy.sparse.linalg import factorized, spsolve
from scipy.sparse import issparse, csc_matrix as sparse

# just in time compiler
# from numba import jit

# Set the complex precision to use
complex_type = complex128


class NodeType(Enum):
    PQ = 1,
    PV = 2,
    REF = 3,
    NONE = 4,
    STO_DISPATCH = 5  # Storage dispatch, in practice it is the same as REF


# @jit(cache=True)
def pre_process(n_bus, Yseries, Vset, pq, pv, vd):
    """
    Make the Helm System matrix
    @param n_bus: Number of buses of the circuit
    @param Yseries: Circuit admittance matrix of the series elements
    @param Vset: Vector of voltages of those nodes where the voltage is controlled (AKA Slack and PV buses)
    @param S: Vector of power injections at all the nodes
    @param pq: list of PQ node indices
    @param pv: list of PV node indices
    @param vd: list of Slack node indices
    @return: 
    """
    
    
    """
    Reduction of the circuit magnitudes.

    Args:
        n_bus: 

        Yseries: 

        slack_indices: Array of indices of the slack nodes

        Vset: 

        S: 

    Output:
        Yred: Reduced admittance matrix (Without the rows and columns belonging to slack buses)

        I: Matrix of currents (In practice only one slack bus is selected, hence it is a vector) injected by the slack buses

        Sred: Array of power injections of the buses that are not of type slack

        types_red: Array of types of the buses that are not of type slack

        non_slack_indices: Array of indices of the buses that are not of type slack
    """

    # now to have efficient arrays of coefficients
    map_idx = zeros(n_bus, dtype=np.int)
    map_w = zeros(n_bus, dtype=np.int)
    npq = 0
    npv = 0
    npqpv = 0

    for i in pq:
        map_idx[i] = npq
        map_w[i] = npqpv
        npq += 1
        npqpv += 1

    for i in pv:
        map_idx[i] = npv
        map_w[i] = npqpv
        npv += 1
        npqpv += 1

    # build the expanded system matrix
    Ysys = zeros((2*n_bus, 2*n_bus))

    for a, b in product(range(n_bus), range(n_bus)):
        Ysys[2*a, 2*b] = Yseries[a, b].real
        Ysys[2*a, 2*b+1] = -Yseries[a, b].imag
        Ysys[2*a+1, 2*b] = Yseries[a, b].imag
        Ysys[2*a+1, 2*b+1] = Yseries[a, b].real

    # set pv column
    for a in pv:
        b = a
        Ysys[:, 2*b] = zeros(2 * n_bus)
        # Ysys[a*2, b*2+1] = 0
        Ysys[a*2+1, b*2] = 1

    # set vd elements
    for a in vd:
        Ysys[a*2, :] = zeros(2 * n_bus)
        Ysys[a*2 + 1, :] = zeros(2 * n_bus)
        Ysys[a*2, a*2] = 1
        Ysys[a*2+1, a*2+1] = 1
    # print('Ysys\n', Ysys)

    # build the PV matrix
    Ypv = zeros((2 * n_bus, npv))
    for a, b in product(r_[pq, pv], pv):
        kk = map_idx[b]
        Ypv[2*a, kk] = Yseries[a, b].real
        Ypv[2*a+1, kk] = Yseries[a, b].imag
    # print('Ypv\n', Ypv)

    Vset2 = Vset * Vset

    return sparse(Ysys), Ypv, Vset2, map_idx, map_w, npq, npv


# @jit(cache=True)
def RHS(n, nbus, Ysh, Ypv, S, Vset, Vset_abs2, C, W, Q, pq, pv, vd, map_idx, map_w):
    """
    Right hand side calculation.

    Args:
        n: Order of the coefficients

        nbus: Number of buses (not counting the slack buses)

        Yrow: Vector where every elements the sum of the corresponding row of the reduced admittance matrix)

        I: Vector of current injections (nbus elements)

        S: Vector of power injections (nbus elements)

        Vset: Vector of set voltages

        Vset_abs2: Vetor with the voltage set points squared. (nbus elements)

        C: Voltage coefficients (Ncoeff x nbus elements)

        X: Weighted coefficients (Ncoeff x nbus elements)

        R: Voltage convolution coefficients (Ncoeff x nbus elements)

        W: Inverse coefficients structure (Ncoeff x nbus elements)

        types: Types of the non slack buses (nbus elements)

    Output:
        rhs: Right hand side vector to solve the coefficients (2 * nbus elements)
    """

    rhs = np.empty(2 * nbus)
    Vre = ones(len(pv))

    for k in pq:
        val = RHS_PQ(n, k, Ysh, C, S, W, map_w)
        rhs[2 * k] = val.real
        rhs[2 * k + 1] = val.imag

    for k in vd:
        val = RHS_VD(n, k, Vset)
        rhs[2 * k] = val.real
        rhs[2 * k + 1] = val.imag

    for k in pv:
        val = RHS_PV(n, k, Ysh, S, C, Q, W, map_idx, map_w)
        rhs[2 * k] = val.real
        rhs[2 * k + 1] = val.imag

        kk = map_idx[k]
        Vre[kk] = calc_Vre(n, k, C, Vset_abs2).real
        rhs -= Ypv[:, kk] * Vre[kk]

    return rhs, Vre


def delta(n, k):
    return n == k  # is 1 for n==k, 0 otherwise


# @jit(cache=True)
def RHS_VD(n, k, Vset):
    """
    Right hand side calculation for a PQ bus.

    Args:
        n: Order of the coefficients

        k: Index of the bus

        Vset: set voltage of the node
    Output:
        Right hand side value for slack nodes
    """
    if n == 0:
        return complex_type(1) * delta(n, 0)
    else:
        return (Vset[k] - complex_type(1)) * delta(n, 1)


# @jit(cache=True)
def RHS_PQ(n, k, Ysh, C, S, W, map_w):
    """
    Right hand side calculation for a PQ bus.

    Args:
        n: Order of the coefficients

        k: Index of the bus

        Ysh: Vector where every elements the sum of the corresponding row of the reduced admittance matrix)

        I: Vector of current injections (nbus elements)

        S: Vector of power injections (nbus elements)

        W: Inverse coefficients structure (Ncoeff x nbus elements)
    Output:
        Right hand side value
    """
    if n == 0:
        return 0

    else:
        kw = map_w[k]
        return conj(S[k]) * conj(W[n-1, kw]) - Ysh[k] * C[n - 1, k]  # ASU version


def calc_W(n, k, kw, C, W):
    """
    Calculation of the inverse coefficients W. (only applicable for PQ buses)

    Args:
        n: Order of the coefficients

        k: Index of the bus

        C: Voltage coefficients (Ncoeff x nbus elements)

        W: Inverse coefficients structure (Ncoeff x nbus elements)

    Output:
        Inverse coefficient of order n for the bus k
    """

    if n == 0:
        res = complex_type(1)
    else:
        res = complex_type(0)
        for l in range(n):
            res -= W[l, kw] * C[n-l, k]

    res /= conj(C[0, k])

    return res


def RHS_PV(n, k, Ysh, S, C, Q, W, map_idx, map_w):
    """
    Right hand side calculation for a PQ bus.

    Args:
        n: Order of the coefficients

        k: Index of the bus

        Ysh: Vector where every elements the sum of the corresponding row of the shunt admittance matrix

        S: Vector of power injections (nbus elements)

        C: Voltage coefficients (Ncoeff x nbus elements)

        Q: Reactive power coefficients (Ncoeff x nbus elements)

        W: Inverse coefficients structure (Ncoeff x nbus elements)
    Output:
        Right hand side value for the pv nodes
    """

    if n == 0:
        return 0  # -1j * Q[0, kk] / conj(C[0, k])

    else:
        kk = map_idx[k]
        kw = map_w[k]
        val = complex_type(0)
        for l in range(1, n):   # this includes the n-1
            val += Q[l, kk] * W[n-l, kw].conjugate()

        n1 = n-1
        rhs = S[k].real * W[n1, kw].conjugate() - (1j * val) - Ysh[k] * C[n1, k]

    return rhs


def calc_Vre(n, k, C, Vset_abs2):
    """
    Compute the real part of the voltage for PV ndes
    Args:
        n: order
        k: PV node index
        C: Structure of voltage coefficients
        Vset_abs2: Square of the set voltage module
    Returns:
        Real part of the voltage for the PV nodes

    """
    
    # vre = delta(n, 0) + 0.5 * delta(n, 1) * (Vset_abs2[k] - 1) - 0.5 * R

    if n == 0:
        return complex_type(1)
    elif n == 1:
        R = calc_R(n, k, C)
        return 0.5 * (Vset_abs2[k] - 1) - 0.5 * R
    else:
        return complex_type(0)

    return vre


def calc_R(n, k, C):
    """
    Convolution coefficient

    Args:
        n: Order of the coefficients

        k: Index of the bus

        C: Voltage coefficients (Ncoeff x nbus elements)

    Output:
        Convolution coefficient of order n for the bus k
    """

    result = complex_type(0)
    for l in range(n+1):
        result += C[l, k] * C[n-l, k].conjugate()

    return result


def epsilon(Sn, n, E):
    """
    Fast recursive Wynn's epsilon algorithm from:
        NONLINEAR SEQUENCE TRANSFORMATIONS FOR THE ACCELERATION OF CONVERGENCE
        AND THE SUMMATION OF DIVERGENT SERIES

        by Ernst Joachim Weniger
    """
    Zero = complex_type(0)
    One = complex_type(1)
    Tiny = np.finfo(complex_type).min
    Huge = np.finfo(complex_type).max

    E[n] = Sn

    if n == 0:
        estim = Sn
    else:
        AUX2 = Zero

        for j in range(n, 0, -1):  # range from n to 1 (both included)
            AUX1 = AUX2
            AUX2 = E[j-1]
            DIFF = E[j] - AUX2

            if abs(DIFF) <= Tiny:
                E[j-1] = Huge
            else:
                if DIFF == 0:
                    DIFF = Tiny
                E[j-1] = AUX1 + One / DIFF

        if mod(n, 2) == 0:
            estim = E[0]
        else:
            estim = E[1]

    return estim, E


def pade_approximation(n, an, s=1):
    """
    Computes the n/2 pade approximant of the series an at the approximation
    point s

    Arguments:
        an: coefficient series
        n:  order of the series
        s: point of approximation

    Returns:
        pade approximation at s
    """
    nn = int(n/2)
    if mod(nn, 2) == 0:
        nn -= 1

    L = nn
    M = nn

    an = np.ndarray.flatten(an)
    rhs = an[L+1:L+M+1]

    C = zeros((L, M), dtype=complex_type)
    for i in range(L):
        k = i + 1
        C[i, :] = an[L-M+k:L+k]
    try:
        b = solve(C, -rhs)  # bn to b1
    except:
        print()
        return 0, zeros(L+1, dtype=complex_type), zeros(L+1, dtype=complex_type)
    b = r_[1, b[::-1]]  # b0 = 1

    a = zeros(L+1, dtype=complex_type)
    a[0] = an[0]
    for i in range(L):
        val = complex_type(0)
        k = i + 1
        for j in range(k+1):
            val += an[k-j] * b[j]
        a[i+1] = val

    p = complex_type(0)
    q = complex_type(0)
    for i in range(L+1):
        p += a[i] * s**i
        q += b[i] * s**i

    return p/q, a, b


def interprete_solution(nbus, npv, pv, pqvd, x_sol, Vre, map_idx):
    """
    Assign the solution vector individual values to the correct places
    Args:
        nbus: number of system nodes
        npv: number of pv nodes
        types: types of each node
        x_sol: solution vector to analyze
        Vre: Vector or real part of the voltage for the PV nodes
        map_idx: mapping array from normal bus index to PV index

    Returns:
        Voltages coefficients and reactive power coefficients for the PV nodes at the order of x_sol
    """
    C = zeros(nbus, dtype=complex_type)
    Q = zeros(npv)

    # for k in pqvd:  # non vectorized code
    #     C[k] = x_sol[2 * k] + 1j * x_sol[2 * k + 1]

    # set the PQ and Slack nodes
    C[pqvd] = x_sol[2 * pqvd] + 1j * x_sol[2 * pqvd + 1]

    # for k in pv:  # non vectorized code
    #     kk = map_idx[k]
    #     Q[kk] = x_sol[2 * k]
    #     C[k] = Vre[kk] + 1j * x_sol[2 * k + 1]

    # Set the PV nodes
    kk = map_idx[pv]
    Q[kk] = x_sol[2 * pv]
    C[pv] = Vre[kk] + 1j * x_sol[2 * pv + 1]

    return C, Q


def helm(Y, Ys, Ysh, max_coefficient_count, S, voltage_set_points, pq, pv, vd, eps=1e-3, use_pade=True):
    """
    Run the holomorphic embedding power flow
    @param Y: Circuit complete admittance matrix
    @param Ys: Circuit series elements admittance matrix
    @param Ysh: Circuit shunt elements admittance matrix
    @param max_coefficient_count: Maximum number of voltage coefficients to evaluate (Must be an odd number)
    @param S: Array of power injections matching the admittance matrix size
    @param voltage_set_points: Array of voltage set points matching the admittance matrix size
    @param pq: list of PQ node indices
    @param pv: list of PV node indices
    @param vd: list of Slack node indices
    @param eps: Tolerance
    @param use_pade: Use the Padè approximation? If False the Epsilon algorithm is used
    @return:
    """

    nbus = np.shape(Ys)[0]

    # The routines in this script are meant to handle sparse matrices, hence non-sparse ones are not allowed
    assert(issparse(Ys))
    # assert(not np.all((Ys + sparse(np.eye(nbus) * Ysh) != Y).data))

    # Make bus type lists combinations that are going to be used later
    pqvd = r_[pq, vd]
    pqvd.sort()
    pqpv = r_[pq, pv]
    pqpv.sort()

    print('Ymat:\n', Y.todense())
    print('Yseries:\n', Ys.todense())
    print('Yshunt:\n', Ysh)

    # prepare the arrays
    Ysys, Ypv, Vset, map_idx, map_w, npq, npv = pre_process(n_bus=nbus, Yseries=Ys, Vset=voltage_set_points,
                                                            pq=pq, pv=pv, vd=vd)

    print('Ysys:\n', Ysys.todense())

    # F = np.zeros(nbus, dtype=complex_type)
    # F[Ysh.indices] = Ysh.data

    # declare the matrix of coefficients that will lead to the voltage computation
    C = zeros((0, nbus), dtype=complex_type)

    # auxiliary array for the epsilon algorithm
    E_v = zeros((0, nbus), dtype=complex_type)
    E_q = zeros((0, npv), dtype=complex_type)

    # Declare the inverse coefficients vector
    # (it is actually a matrix; a vector of coefficients per coefficient order)
    W = zeros((0, npq+npv), dtype=complex_type)

    # Reactive power on the PV nodes
    Q = zeros((0, npv), dtype=complex_type)

    # Squared values of the voltage module for the buses that are not of slack type
    Vset_abs2 = abs(voltage_set_points) ** 2

    # progressive calculation of coefficients
    n = 0
    converged = False
    inside_precision = True
    errors = list()
    errors_PV_P = list()
    errors_PV_Q = list()
    errors_PQ_P= list()
    errors_PQ_Q = list()
    voltages = list()
    Sn_v = zeros(nbus, dtype=complex_type)
    Sn_q = zeros(npv, dtype=complex_type)
    voltages_vector = zeros(nbus, dtype=complex_type)
    Vred_last = zeros(nbus, dtype=complex_type)
    solve = factorized(Ysys)

    # set the slack indices voltages
    voltages_vector[vd] = voltage_set_points[vd]

    while n <= max_coefficient_count and not converged and inside_precision:

        # Reserve coefficients memory space
        C = np.vstack((C, np.zeros((1, nbus), dtype=complex_type)))
        E_v = np.vstack((E_v, np.zeros((1, nbus), dtype=complex_type)))
        E_q = np.vstack((E_q, np.zeros((1, npv))))
        W = np.vstack((W, np.zeros((1, npq+npv), dtype=complex_type)))
        Q = np.vstack((Q, np.zeros((1, npv), dtype=complex_type)))

        # get the system independent term to solve the coefficients
        # n, nbus, F, Ypv, S, Vset, Vset_abs2, C, W, Q, pq, pv, vd, map_idx, map_w,
        rhs, Vre = RHS(n, nbus, Ysh, Ypv, S, Vset, Vset_abs2, C, W, Q, pq, pv, vd, map_idx, map_w)

        # Solve the linear system to obtain the new coefficients
        x_sol = solve(rhs)

        # assign the voltages and the reactive power values correctly
        C[n, :], Q[n, :] = interprete_solution(nbus, npv, pv, pqvd, x_sol, Vre, map_idx)

        # copy variables for the epsilon algorithm
        if not use_pade:
            E_v[n, :] = C[n, :]
            E_q[n, :] = Q[n, :]
            Sn_v += C[n, :]
            Sn_q += Q[n, :]

        # Update the inverse voltage coefficients W for the non slack nodes
        for k in pqpv:
            kw = map_w[k]  # actual index in the coefficients structure
            W[n, kw] = calc_W(n, k, kw, C, W)

        # calculate the reactive power
        for k in pv:
            kk = map_idx[k]
            if use_pade:
                if mod(n, 2) == 0 and n > 2:
                    q, _, _ = pade_approximation(n, Q)
                    S[k] = S[k].real + 1j * q.real
            else:
                q, E_q[:, kk] = epsilon(Sn_q[kk], n, E_q[:, kk])
                S[k] = S[k].real + 1j * q

        # calculate the voltages
        for k in pqpv:
            if use_pade:
                if mod(n, 2) == 0 and n > 2:
                    v, _, _ = pade_approximation(n, C[:, k])
                    voltages_vector[k] = v
            else:
                voltages_vector[k], E_v[:, k] = epsilon(Sn_v[k], n, E_v[:, k])

            if np.isnan(voltages_vector[k]):
                print('Maximum precision reached at ', n)
                voltages_vector = Vred_last
                inside_precision = False
                break

        Vred_last = voltages_vector.copy()

        # Compose the voltage values from the coefficient series
        voltages.append(voltages_vector.copy())
        # print(voltages_vector)

        # Calculate the error and check the convergence
        Scalc = voltages_vector * conj(Y * voltages_vector)
        power_mismatch = Scalc - S  # complex power mismatch
        power_mismatch_ = r_[power_mismatch[pv].real, power_mismatch[pq].real, power_mismatch[pq].imag]  # concatenate error by type

        # check for convergence
        normF = linalg.norm(power_mismatch_, Inf)
        errors.append(normF)
        errors_PV_P.append(power_mismatch[pv].real)
        errors_PV_Q.append(power_mismatch[pv].imag)
        errors_PQ_P.append(power_mismatch[pq].real)
        errors_PQ_Q.append(power_mismatch[pq].imag)
        if normF < eps:
            converged = True
        else:
            converged = False

        n += 1  # increase the coefficients order

    # errors_lst = [array(errors), array(errors_PV_P), array(errors_PV_Q), array(errors_PQ_P), array(errors_PQ_Q)]

    return Vred_last, converged, normF, Scalc


def bifurcation_point(C, slackIndices):
    """
    Computes the bifurcation point
    @param C:
    @return:
    """
    npoints = 100
    order_num, bus_num = np.shape(C)
    # V(S) = P(S)/Q(S)
    V = zeros((npoints, bus_num), dtype=complex_type)
    L = zeros((npoints, bus_num))
    for k in range(bus_num):
        if k not in slackIndices:
            _, p, q = pade_approximation(order_num, C[:, k])
            # print(k, 'P:', p)
            # print(k, 'Q:', q)
            asint = np.roots(q[::-1])
            asint = np.sort(abs(asint))
            asint = asint[asint > 2]
            print('Asymptotes', asint)
            # print('Asymptote:', asint[0])

            bpoint = asint[0]
            # bpoint = max(asint)
            lmda = np.linspace(1, bpoint, npoints+1)[:-1]
            # lmda = np.linspace(1, 100, npoints)

            pval = np.polyval(p[::-1], lmda)
            qval = np.polyval(q[::-1], lmda)

            V[:, k] = pval / qval
            L[:, k] = lmda

    return V, L





import numpy as np


def thevenin_funcX2(U, X):
    n = len(U)
    r_3 = np. zeros(n, dtype=complex)
    r_2 = np. zeros(n, dtype=complex)
    r_1 = np. zeros(n, dtype=complex)
    r_0 = np. zeros(n, dtype=complex)
    T_03 = np. zeros(n, dtype=complex)
    T_02 = np. zeros(n, dtype=complex)
    T_01 = np. zeros(n, dtype=complex)
    T_00 = np. zeros(n, dtype=complex)
    T_13 = np. zeros(n, dtype=complex)
    T_12 = np. zeros(n, dtype=complex)
    T_11 = np. zeros(n, dtype=complex)
    T_10 = np. zeros(n, dtype=complex)
    T_23 = np. zeros(n, dtype=complex)
    T_22 = np. zeros(n, dtype=complex)
    T_21 = np. zeros(n, dtype=complex)
    T_20 = np. zeros(n, dtype=complex)

    r_0[0] = -1
    r_1[0:n-1] = U[1:n]
    r_2[0:n-2] = U[2:n] - U[1] * X[1:n-1]

    T_00[0] = -1
    T_01[0] = -1
    T_02[0] = -1
    T_10[0] = 0
    T_11[0] = 1
    T_12[0] = 1
    T_20[0] = 0
    T_21[0] = 0
    T_22[0] = -U[1]

    for l in range(n):  # ANAR CALCULANT CONSTANTS , RESIDUS I POLINOMIS
        a = (r_2[0] * r_1[0]) / (- r_0[1] * r_1[0] + r_0[0] * r_1[1] - r_0[0] * r_2[0])
        b = -a * r_0[0] / r_1[0]
        c = 1 - b
        T_03[0] = b * T_01[0] + c * T_02[0]
        T_03[1:n] = a * T_00[0:n-1] + b * T_01[1:n] + c * T_02[1:n]
        T_13[0] = b * T_11[0] + c * T_12[0]
        T_13[1:n] = a * T_10[0:n-1] + b * T_11[1:n] + c * T_12[1:n]
        T_23[0] = b * T_21[0] + c * T_22[0]
        T_23[1:n] = a * T_20[0:n-1] + b * T_21[1:n] + c * T_22[1:n]
        r_3[0:n-2] = a * r_0[2:n] + b * r_1[2:n] + c * r_2[1:n-1]

        if l == n - 1:
            t_0 = T_03
            t_1 = T_13
            t_2 = T_23

        r_0[:] = r_1[:]
        r_1[:] = r_2[:]
        r_2[:] = r_3[:]
        T_00[:] = T_01[:]
        T_01[:] = T_02[:]
        T_02[:] = T_03[:]
        T_10[:] = T_11[:]
        T_11[:] = T_12[:]
        T_12[:] = T_13[:]
        T_20[:] = T_21[:]
        T_21[:] = T_22[:]
        T_22[:] = T_23[:]

        r_3 = np.zeros(n, dtype=complex)
        T_03 = np.zeros(n, dtype=complex)
        T_13 = np.zeros(n, dtype=complex)
        T_23 = np.zeros(n, dtype=complex)

    usw = -sum(t_0) / sum(t_1)
    sth = -sum(t_2) / sum(t_1)

    sigma_bo = sth / (usw * np.conj(usw))

    u = 0.5 + np.sqrt(0.25 + np.real(sigma_bo) - np. imag(sigma_bo)**2) + np.imag(sigma_bo)*1j  # positive branch
    #u = 0.5 - np.sqrt(0.25 + np.real(sigma_bo) - np.imag(sigma_bo) ** 2) + np.imag(sigma_bo) * 1j  # negative branch
    ufinal = u*usw

    return ufinal


def thevenin4all(U, X):
    """

    :param U:
    :param X:
    :return:
    """
    n_coeff, nbus = U.shape
    r_3 = np. zeros((n_coeff, nbus), dtype=complex)
    r_2 = np. zeros((n_coeff, nbus), dtype=complex)
    r_1 = np. zeros((n_coeff, nbus), dtype=complex)
    r_0 = np. zeros((n_coeff, nbus), dtype=complex)
    T_03 = np. zeros((n_coeff, nbus), dtype=complex)
    T_02 = np. zeros((n_coeff, nbus), dtype=complex)
    T_01 = np. zeros((n_coeff, nbus), dtype=complex)
    T_00 = np. zeros((n_coeff, nbus), dtype=complex)
    T_13 = np. zeros((n_coeff, nbus), dtype=complex)
    T_12 = np. zeros((n_coeff, nbus), dtype=complex)
    T_11 = np. zeros((n_coeff, nbus), dtype=complex)
    T_10 = np. zeros((n_coeff, nbus), dtype=complex)
    T_23 = np. zeros((n_coeff, nbus), dtype=complex)
    T_22 = np. zeros((n_coeff, nbus), dtype=complex)
    T_21 = np. zeros((n_coeff, nbus), dtype=complex)
    T_20 = np. zeros((n_coeff, nbus), dtype=complex)

    r_0[0, :] = -1.0
    r_1[0:n_coeff-1, :] = U[1:n_coeff, :]
    r_2[0:n_coeff-2, :] = U[2:n_coeff, :] - U[1, :] * X[1:n_coeff-1, :]

    T_00[0, :] = -1.0
    T_01[0, :] = -1.0
    T_02[0, :] = -1.0
    T_10[0, :] = 0.0
    T_11[0, :] = 1.0
    T_12[0, :] = 1.0
    T_20[0, :] = 0.0
    T_21[0, :] = 0.0
    T_22[0, :] = -U[1, :]

    for l in range(n_coeff):  # ANAR CALCULANT CONSTANTS , RESIDUS I POLINOMIS

        a = (r_2[0, :] * r_1[0, :]) / (- r_0[1, :] * r_1[0, :] + r_0[0, :] * r_1[1, :] - r_0[0, :] * r_2[0, :])
        b = -a * r_0[0, :] / r_1[0, :]
        c = 1.0 - b
        T_03[0, :] = b * T_01[0, :] + c * T_02[0, :]
        T_03[1:n_coeff, :] = a * T_00[0:n_coeff-1, :] + b * T_01[1:n_coeff, :] + c * T_02[1:n_coeff, :]
        T_13[0, :] = b * T_11[0, :] + c * T_12[0, :]
        T_13[1:n_coeff, :] = a * T_10[0:n_coeff-1, :] + b * T_11[1:n_coeff, :] + c * T_12[1:n_coeff, :]
        T_23[0, :] = b * T_21[0, :] + c * T_22[0, :]
        T_23[1:n_coeff, :] = a * T_20[0:n_coeff-1, :] + b * T_21[1:n_coeff, :] + c * T_22[1:n_coeff, :]
        r_3[0:n_coeff-2, :] = a * r_0[2:n_coeff, :] + b * r_1[2:n_coeff, :] + c * r_2[1:n_coeff-1, :]

        if l == n_coeff - 1:
            t_0 = T_03
            t_1 = T_13
            t_2 = T_23

        r_0 = r_1
        r_1 = r_2
        r_2 = r_3
        T_00 = T_01
        T_01 = T_02
        T_02 = T_03
        T_10 = T_11
        T_11 = T_12
        T_12 = T_13
        T_20 = T_21
        T_21 = T_22
        T_22 = T_23

        r_3 = np.zeros((n_coeff, nbus), dtype=complex)
        T_03 = np.zeros((n_coeff, nbus), dtype=complex)
        T_13 = np.zeros((n_coeff, nbus), dtype=complex)
        T_23 = np.zeros((n_coeff, nbus), dtype=complex)

    usw = -t_0.sum(axis=0) / t_1.sum(axis=0)
    sth = -t_2.sum(axis=0) / t_1.sum(axis=0)

    sigma_bo = sth / (usw * np.conj(usw))

    u = 0.5 + np.sqrt(0.25 + np.real(sigma_bo) - np. imag(sigma_bo)**2) + np.imag(sigma_bo) * 1j  # positive branch
    # u = 0.5 - np.sqrt(0.25 + np.real(sigma_bo) - np.imag(sigma_bo) ** 2) + np.imag(sigma_bo) * 1j  # negative branch
    ufinal = u * usw

    return ufinal


if __name__ == '__main__':
    from GridCal.Engine import FileOpen
    from GridCal.Engine.Simulations.PowerFlow.helm_power_flow import helm_coefficients_josep, pade4all
    import pandas as pd

    np.set_printoptions(linewidth=2000, suppress=True)
    pd.set_option('display.max_rows', 500)
    pd.set_option('display.max_columns', 500)
    pd.set_option('display.width', 1000)

    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/IEEE39_1W.gridcal'
    fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/IEEE 14.xlsx'
    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/lynn5buspv.xlsx'
    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/IEEE 118.xlsx'
    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/1354 Pegase.xlsx'
    # fname = 'helm_data1.gridcal'
    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/IEEE 14 PQ only.gridcal'
    # fname = 'IEEE 14 PQ only full.gridcal'
    grid = FileOpen(fname).open()

    nc = grid.compile_snapshot()
    inputs = nc.compute()[0]  # pick the first island

    U_, X_, Q_, iter_ = helm_coefficients_josep(Yseries=inputs.Yseries,
                                                V0=inputs.Vbus,
                                                S0=inputs.Sbus,
                                                Ysh0=inputs.Ysh,
                                                pq=inputs.pq,
                                                pv=inputs.pv,
                                                sl=inputs.ref,
                                                pqpv=inputs.pqpv,
                                                tolerance=1e-6,
                                                max_coeff=10,
                                                verbose=False)

    # get the voltages with the thevenin approximant
    thevenin_voltage = thevenin4all(U=U_, X=X_)
    pade_voltage = pade4all(order=U_.shape[0]-1, coeff_mat=U_, s=1)

    # check the implementation of thevenin4all
    n = U_.shape[1]
    for i in range(n):
        vsel = thevenin_funcX2(U=U_[:, i], X=X_[:, i])
        assert np.isclose(thevenin_voltage[i], vsel)

    data = np.c_[np.abs(U_.sum(axis=0)), np.abs(pade_voltage), np.abs(thevenin_voltage)]
    df = pd.DataFrame(data=data, columns=['Sum', 'Padè', 'Thevenin'])
    print(df)


import numpy as np
np.set_printoptions(precision=6, suppress=True, linewidth=320)
from numpy import where, zeros, ones, mod, conj, array, dot, angle, complex128 #, complex256
from numpy.linalg import solve

# Set the complex precision to use
complex_type = complex128


def calc_W(n, npqpv, C, W):
    """
    Calculation of the inverse coefficients W.
    @param n: Order of the coefficients
    @param npqpv: number of pq and pv nodes
    @param C: Structure of voltage coefficients (Ncoeff x nbus elements)
    @param W: Structure of inverse voltage coefficients (Ncoeff x nbus elements)
    @return: Array of inverse voltage coefficients for the order n
    """

    if n == 0:
        res = ones(npqpv, dtype=complex_type)
    else:
        l = arange(n)
        res = -(W[l, :] * C[n - l, :]).sum(axis=0)

    res /= conj(C[0, :])

    return res


def pade_approximation(n, an, s=1):
    """
    Computes the n/2 pade approximant of the series an at the approximation
    point s
    Arguments:
        an: coefficient matrix, (number of coefficients, number of series)
        n:  order of the series
        s: point of approximation
    Returns:
        pade approximation at s
    """
    nn = int(n / 2)
    if mod(nn, 2) == 0:
        nn -= 1

    L = nn
    M = nn

    an = np.ndarray.flatten(an)
    rhs = an[L + 1:L + M + 1]

    C = zeros((L, M), dtype=complex_type)
    for i in range(L):
        k = i + 1
        C[i, :] = an[L - M + k:L + k]

    try:
        b = solve(C, -rhs)  # bn to b1
    except:
        return 0, zeros(L + 1, dtype=complex_type), zeros(L + 1, dtype=complex_type)

    b = r_[1, b[::-1]]  # b0 = 1

    a = zeros(L + 1, dtype=complex_type)
    a[0] = an[0]
    for i in range(L):
        val = complex_type(0)
        k = i + 1
        for j in range(k + 1):
            val += an[k - j] * b[j]
        a[i + 1] = val

    p = complex_type(0)
    q = complex_type(0)
    for i in range(L + 1):
        p += a[i] * s ** i
        q += b[i] * s ** i

    return p / q, a, b


# @jit(cache=True)
def helmz(Vbus, Sbus, Ibus, Ybus, pq, pv, ref, pqpv, tol=1e-9, max_ter=5):
    """

    Args:
        admittances: Circuit complete admittance matrix

        slackIndices: Indices of the slack buses (although most likely only one works)

        coefficientCount: Number of voltage coefficients to evaluate (Must be an odd number)

        powerInjections: Array of power injections matching the admittance matrix size

        voltageSetPoints: Array of voltage set points matching the admittance matrix size

        types: Array of bus types matching the admittance matrix size. types: {1-> PQ, 2-> PV, 3-> Slack}

    Output:
        Voltages vector
    """

    # reduced impedance matrix
    Zred = inv(Ybus[pqpv, :][:, pqpv]).toarray()

    # slack currents
    Ivd = -Ybus[pqpv, :][:, ref].dot(Vbus[ref])

    # slack voltages influence
    Ck = Zred.dot(Ivd)

    npqpv = len(pqpv)

    Vcoeff = zeros((0, npqpv), dtype=complex_type)
    Wcoeff = zeros((0, npqpv), dtype=complex_type)

    row = zeros((1, npqpv), dtype=complex_type)

    for n in range(max_ter):

        # reserve memory
        Vcoeff = r_[Vcoeff, row.copy()]
        Wcoeff = r_[Wcoeff, row.copy()]

        if n == 0:
            I = Ivd
        else:
            I = conj(Sbus[pqpv]) * Wcoeff[n-1, :]

        # solve the voltage coefficients
        Vcoeff[n, :] = Zred.dot(I)

        # compute the inverse voltage coefficients
        Wcoeff[n, :] = calc_W(n=n, npqpv=npqpv, C=Vcoeff, W=Wcoeff)

    # compose the final voltage
    voltage = Vbus.copy()

    for i, ii in enumerate(pqpv):
        voltage[ii], _, _ = pade_approximation(n, Vcoeff[:, i])

    # evaluate F(x)
    Scalc = voltage * conj(Ybus * voltage - Ibus)
    mis = Scalc - Sbus  # complex power mismatch
    normF = linalg.norm(r_[mis[pv].real, mis[pq].real, mis[pq].imag], Inf)

    print('Vcoeff:\n', Vcoeff)

    print('V:\n', abs(Vcoeff.sum(axis=0)))

    return voltage, normF


if __name__ == "__main__":
    from GridCal.Engine.calculation_engine import *

    grid = MultiCircuit()
    grid.load_file('lynn5buspq.xlsx')

    grid.compile()

    circuit = grid.circuits[0]

    print('\nYbus:\n', circuit.power_flow_input.Ybus.todense())
    print('\nYseries:\n', circuit.power_flow_input.Yseries.todense())
    print('\nYshunt:\n', circuit.power_flow_input.Yshunt)
    print('\nSbus:\n', circuit.power_flow_input.Sbus)
    print('\nIbus:\n', circuit.power_flow_input.Ibus)
    print('\nVbus:\n', circuit.power_flow_input.Vbus)
    print('\ntypes:\n', circuit.power_flow_input.types)
    print('\npq:\n', circuit.power_flow_input.pq)
    print('\npv:\n', circuit.power_flow_input.pv)
    print('\nvd:\n', circuit.power_flow_input.ref)

    import time
    print('HELM-Z')
    start_time = time.time()
    cmax = 40
    V1, err = helmz(Vbus=circuit.power_flow_input.Vbus,
                    Sbus=circuit.power_flow_input.Sbus,
                    Ibus=circuit.power_flow_input.Ibus,
                    Ybus=circuit.power_flow_input.Yseries,
                    pq=circuit.power_flow_input.pq,
                    pv=circuit.power_flow_input.pv,
                    ref=circuit.power_flow_input.ref,
                    pqpv=circuit.power_flow_input.pqpv,
                    max_ter=cmax)

    print("--- %s seconds ---" % (time.time() - start_time))
    # print_coeffs(C, W, R, X, H)

    print('V module:\t', abs(V1))
    print('V angle: \t', angle(V1))
    print('error: \t', err)

    # check the HELM solution: v against the NR power flow
    print('\nNR')
    options = PowerFlowOptions(SolverType.NR, verbose=False, robust=False, tolerance=1e-9)
    power_flow = PowerFlow(grid, options)

    start_time = time.time()
    power_flow.run()
    print("--- %s seconds ---" % (time.time() - start_time))
    vnr = circuit.power_flow_results.voltage

    print('V module:\t', abs(vnr))
    print('V angle: \t', angle(vnr))
    print('error: \t', circuit.power_flow_results.error)

    # check
    print('\ndiff:\t', V1 - vnr)



# This file is part of GridCal.
#
# GridCal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GridCal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GridCal.  If not, see <http://www.gnu.org/licenses/>.
import pandas as pd
from matplotlib import pyplot as plt
from GridCal.Engine.Devices.editable_device import EditableDevice, DeviceType, GCProp


class ExternalGrid(EditableDevice):
    """
    External grid device
    In essence, this is a slack-enforcer device

    Arguments:

        **name** (str, "Load"): Name of the load

        **G** (float, 0.0): Conductance in equivalent MW

        **B** (float, 0.0): Susceptance in equivalent MVAr

        **Ir** (float, 0.0): Real current in equivalent MW

        **Ii** (float, 0.0): Imaginary current in equivalent MVAr

        **P** (float, 0.0): Active power in MW

        **Q** (float, 0.0): Reactive power in MVAr

        **G_prof** (DataFrame, None): Pandas DataFrame with the conductance profile in equivalent MW

        **B_prof** (DataFrame, None): Pandas DataFrame with the susceptance profile in equivalent MVAr

        **Ir_prof** (DataFrame, None): Pandas DataFrame with the real current profile in equivalent MW

        **Ii_prof** (DataFrame, None): Pandas DataFrame with the imaginary current profile in equivalent MVAr

        **P_prof** (DataFrame, None): Pandas DataFrame with the active power profile in equivalent MW

        **Q_prof** (DataFrame, None): Pandas DataFrame with the reactive power profile in equivalent MVAr

        **active** (bool, True): Is the load active?

        **mttf** (float, 0.0): Mean time to failure in hours

        **mttr** (float, 0.0): Mean time to recovery in hours

    """

    def __init__(self, name='External grid', idtag=None, active=True, Vm=1.0, Va=0.0, Vm_prof=None, Va_prof=None,
                  mttf=0.0, mttr=0.0):

        EditableDevice.__init__(self,
                                name=name,
                                idtag=idtag,
                                active=active,
                                device_type=DeviceType.ExternalGridDevice,
                                editable_headers={'name': GCProp('', str, 'Load name'),
                                                  'idtag': GCProp('', str, 'Unique ID'),
                                                  'bus': GCProp('', DeviceType.BusDevice, 'Connection bus name'),
                                                  'active': GCProp('', bool, 'Is the load active?'),
                                                  'Vm': GCProp('p.u.', float, 'Active power'),
                                                  'Va': GCProp('radians', float, 'Reactive power'),
                                                  'mttf': GCProp('h', float, 'Mean time to failure'),
                                                  'mttr': GCProp('h', float, 'Mean time to recovery'),
                                                  },
                                non_editable_attributes=['bus', 'idtag'],
                                properties_with_profile={'active': 'active_prof',
                                                         'Vm': 'Vm_prof',
                                                         'Va': 'Va_prof'})

        self.bus = None

        self.active_prof = None

        self.mttf = mttf

        self.mttr = mttr

        # Impedance in equivalent MVA
        self.Vm = Vm
        self.Va = Va
        self.Vm_prof = Vm_prof
        self.Va_prof = Va_prof

    def copy(self):

        elm = ExternalGrid()

        elm.name = self.name
        elm.active = self.active
        elm.active_prof = self.active_prof
        elm.Vm = self.Vm
        elm.Va = self.Va
        elm.Vm_prof = self.Vm_prof
        elm.Va_prof = self.Va_prof

        elm.mttf = self.mttf
        elm.mttr = self.mttr

        return elm

    def get_properties_dict(self):
        """
        Get json dictionary
        :return:
        """

        d = {'id': self.idtag,
             'type': 'load',
             'phases': 'ps',
             'name': self.name,
             'bus': self.bus.idtag,
             'active': self.active,
             'Vm': self.Vm,
             'Va': self.Va}

        if self.active_prof is not None:
            d['active_profile'] = self.active_prof.tolist()
            d['Vm_prof'] = self.Vm_prof.tolist()
            d['Va_prof'] = self.Va_prof.tolist()

        return d

    def get_profiles_dict(self):
        """

        :return:
        """
        if self.active_prof is not None:
            active_profile = self.active_prof.tolist()
            Vm_prof = self.Vm_prof.tolist()
            Va_prof = self.Va_prof.tolist()
        else:
            active_profile = list()
            Vm_prof = list()
            Va_prof = list()

        return {'id': self.idtag,
                'active': active_profile,
                'vm': Vm_prof,
                'va': Va_prof}

    def plot_profiles(self, time=None, show_fig=True):
        """
        Plot the time series results of this object
        :param time: array of time values
        :param show_fig: Show the figure?
        """

        if time is not None:
            fig = plt.figure(figsize=(12, 8))

            ax_1 = fig.add_subplot(211)
            ax_2 = fig.add_subplot(212)

            # P
            y = self.Vm_prof
            df = pd.DataFrame(data=y, index=time, columns=[self.name])
            ax_1.set_title('Voltage module', fontsize=14)
            ax_1.set_ylabel('p.u.', fontsize=11)
            df.plot(ax=ax_1)

            # Q
            y = self.Va_prof
            df = pd.DataFrame(data=y, index=time, columns=[self.name])
            ax_2.set_title('Voltage angle', fontsize=14)
            ax_2.set_ylabel('radians', fontsize=11)
            df.plot(ax=ax_2)

            plt.legend()
            fig.suptitle(self.name, fontsize=20)

            if show_fig:
                plt.show()



"""
.. module:: strategy
   :synopsis: Basic strategy classes for asynchronous optimization.
.. moduleauthor:: David Bindel <bindel@cornell.edu>
"""

try:
    import Queue
except ImportError:
    import queue as Queue

import copy
import threading
import random
import logging
from collections import deque


# Get module-level logger
logger = logging.getLogger(__name__)


class Proposal(object):
    """Represent a proposed action.

    We currently recognize three types of proposals: evaluation requests
    ("eval"), requests to stop an evaluation ("kill"), and requests
    to terminate the optimization ("terminate").  When an evaluation
    proposal is accepted, the controller adds an evaluation record to the
    proposal before notifying subscribers.

    NB: Callbacks that handle proposal accept/reject should be
    insensitive to order: it is important *not* to modify attributes
    that might be used by later callbacks (e.g. the proposal.accepted
    flag).

    Attributes:
        action: String describing the action
        args: Tuple of arguments to pass to the action
        accepted: Flag set on accept/reject decision
        callbacks: Functions to call on accept/reject of action
        record: Evaluation record for accepted evaluation proposals

    """

    def __init__(self, action, *args):
        """Initialize the request.

        Args:
            action: String describing the action
            args:   Tuple of arguments to pass to the action
        """
        self.action = action
        self.args = args
        self.accepted = False
        self.callbacks = []

    def copy(self):
        "Make a copy of the proposal."
        proposal = copy.copy(self)
        proposal.callbacks = copy.copy(self.callbacks)
        return proposal

    def add_callback(self, callback):
        "Add a callback subscriber to the action."
        self.callbacks.append(callback)

    def remove_callback(self, callback):
        "Remove a callback subscriber."
        self.callbacks = [c for c in self.callbacks if c != callback]

    def accept(self):
        "Mark the action as accepted and execute callbacks."
        logger.debug("Running callbacks on proposal accept")
        self.accepted = True
        for callback in self.callbacks:
            callback(self)

    def reject(self):
        "Mark action as rejected and execute callbacks."
        logger.debug("Running callbacks on proposal reject")
        self.accepted = False
        for callback in self.callbacks:
            callback(self)


class EvalRecord(object):
    """Represent progress of a function evaluation.

    An evaluation record includes the evaluation point, status of the
    evaluation, and a list of callbacks.  The status may be pending,
    running, killed (deliberately), cancelled (due to a crash or
    failure) or completed.  Once the function evaluation is completed,
    the value is stored as an attribute in the evaluation record.
    Other information, such as gradients, Hessians, or function values
    used to compute constraints, may also be stored in the record.

    The evaluation record callbacks are triggered on any relevant
    event, including not only status changes but also intermediate
    results.  The nature of these intermediate results (lower bounds,
    initial estimates, etc) is somewhat application-dependent.

    NB: Callbacks that handle record updates should be insensitive to
    order: it is important *not* to modify attributes that might be
    used by later callbacks (e.g. the proposal.accepted flag).

    Attributes:
        params: Evaluation point for the function
        status: Status of the evaluation (pending, running, killed, completed)
        value: Return value (if completed)
        callbacks: Functions to call on status updates

    """

    def __init__(self, params, extra_args=None, status='pending'):
        """Initialize the record.

        Args:
            params: Evaluation point for the function
            extra_args: Extra arguments to be sent to the worker
        Kwargs:
            status: Status of the evaluation (default 'pending')
        """
        self.params = params
        self.extra_args = extra_args
        self._status = status
        self.value = None
        self.callbacks = []

    def add_callback(self, callback):
        "Add a callback for update events."
        self.callbacks.append(callback)

    def remove_callback(self, callback):
        "Remove a callback subscriber."
        self.callbacks = [c for c in self.callbacks if c != callback]

    def update(self, **kwargs):
        "Update fields and execute callbacks."
        logger.debug("Running callbacks on feval update")
        if kwargs is not None:
            for key, value in kwargs.items():
                setattr(self, key, value)
        for callback in self.callbacks:
            callback(self)

    def update_dict(self, dict):
        """Update fields and execute callbacks.

        Args:
            dict: Dictionary of attributes to set on the record
        """
        self.update(**dict)

    def running(self):
        "Change status to 'running' and execute callbacks."
        if self.is_done:
            logger.error("Cannot update record that is done [running]")
        assert not self.is_done, "Cannot change complete to running"
        self.update(_status='running')

    def kill(self):
        "Change status to 'killed' and execute callbacks."
        if self.is_done:
            logger.error("Cannot update record that is done [kill]")
        assert not self.is_done, "Cannot change complete to killed"
        self.update(_status='killed')

    def cancel(self):
        "Change status to 'cancelled' and execute callbacks."
        if self.is_done:
            logger.error("Cannot update record that is done [cancel]")
        assert not self.is_done, "Cannot change complete to cancelled"
        self.update(_status='cancelled')

    def complete(self, value):
        "Change status to 'completed' and execute callbacks."
        if self.is_done:
            logger.error("Cannot update record that is done [complete]")
        assert not self.is_done, "Cannot re-complete"
        self.update(_status='completed', value=value)

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, v):
        assert not self.is_done, "Cannot complete already-completed record"

    @property
    def is_pending(self):
        "Check if status is pending."
        return self._status == 'pending'

    @property
    def is_running(self):
        "Check if status is running."
        return self._status == 'running'

    @property
    def is_completed(self):
        "Check for successful completion of evaluation"
        return self._status == 'completed'

    @property
    def is_killed(self):
        "Check whether evaluation is killed"
        return self._status == 'killed'

    @property
    def is_cancelled(self):
        "Check whether evaluation was cancelled"
        return self._status == 'cancelled'

    @property
    def is_done(self):
        "Check whether the status indicates the evaluation is finished."
        return (self._status == 'completed' or
                self._status == 'killed' or
                self._status == 'cancelled')


class BaseStrategy(object):
    """Base strategy class.

    The BaseStrategy class provides some support for the basic callback
    flow common to most strategies: handlers are called when an evaluation
    is accepted or rejected; and if an evaluation proposal is accepted,
    the record is decorated so that an on_update handler is called for
    subsequent updates.

    Note that not all strategies follow this pattern -- the only requirement
    for a strategy is that it implement propose_action.
    """

    def propose_eval(self, *args):
        "Generate an eval proposal with a callback to on_reply."
        proposal = Proposal('eval', *args)
        proposal.add_callback(self.on_reply)
        return proposal

    def propose_kill(self, r):
        "Generate a kill proposal with a callback to on_reply_kill."
        proposal = Proposal('kill', r)
        proposal.add_callback(self.on_kill_reply)
        return proposal

    def propose_terminate(self):
        "Generate a terminate proposal with a callback to on_terminate."
        proposal = Proposal('terminate')
        proposal.add_callback(self.on_terminate_reply)
        return proposal

    def decorate_proposal(self, proposal):
        "Add a callback to an existing proposal (for nested strategies)."
        if proposal is None:
            return proposal
        elif proposal.action == 'eval':
            proposal.add_callback(self.on_reply)
        elif proposal.action == 'kill':
            proposal.add_callback(self.on_kill_reply)
        elif proposal.action == 'terminate':
            proposal.add_callback(self.on_terminate_reply)
        return proposal

    def on_reply(self, proposal):
        "Default handling of eval proposal."
        logger.debug("Receive proposal reply at base strategy")
        if proposal.accepted:
            logger.debug("Add on_update callback to record")
            proposal.record.add_callback(self.on_update)
            self.on_reply_accept(proposal)
        else:
            self.on_reply_reject(proposal)

    def on_reply_accept(self, proposal):
        "Handle proposal acceptance."
        pass

    def on_reply_reject(self, proposal):
        "Handle proposal rejection."
        pass

    def on_kill_reply(self, proposal):
        "Default handling of kill proposal."
        logger.debug("Received kill proposal reply at base strategy")
        if proposal.accepted:
            self.on_kill_reply_accept(proposal)
        else:
            self.on_kill_reply_reject(proposal)

    def on_kill_reply_accept(self, proposal):
        "Handle proposal acceptance."
        pass

    def on_kill_reply_reject(self, proposal):
        "Handle proposal rejection."
        pass

    def on_terminate_reply(self, proposal):
        "Default handling of terminate proposal."
        logger.debug("Received terminate proposal reply at base strategy")
        if proposal.accepted:
            self.on_terminate_reply_accept(proposal)
        else:
            self.on_terminate_reply_reject(proposal)

    def on_terminate_reply_accept(self, proposal):
        "Handle proposal acceptance."
        pass

    def on_terminate_reply_reject(self, proposal):
        "Handle proposal rejection."
        pass

    def on_update(self, record):
        "Process update."
        if record.is_completed:
            logger.debug("Received record completed at base strategy")
            self.on_complete(record)
        elif record.is_done:
            logger.debug("Received record killed at base strategy")
            self.on_kill(record)
        else:
            logger.debug("Intermediate record update at base strategy")

    def on_complete(self, record):
        "Process completed record."
        pass

    def on_kill(self, record):
        "Process killed or cancelled record."
        pass


class RetryStrategy(BaseStrategy):
    """Retry strategy class.

    The RetryStrategy class manages a queue of proposals to be retried,
    either because they were rejected or because they correspond to
    a function evaluation that was killed.

    If a proposal is submitted to a retry strategy and it has already
    been marked with the .retry attribute, we do not attempt to manage
    retries, as another object is assumed to have taken responsibility
    for retrying the proposal on failure.  This prevents us from having
    redundant retries.

    When a proposal is submitted to the retry class, we make a copy
    (the proposal draft) for safe-keeping.  If at any point it is
    necessary to resubmit, we submit the draft copy to the retry
    class.  The draft will have any modifications or callbacks that
    were added before the proposal reached this strategy, and none of
    those that were added afterward (including those added afterward
    by the RetryStrategy itself).  The intent is that retried
    proposals inserted into the strategy hierarchy at the place where
    they were recorded should require no special treatment to avoid
    adding multiple copies of callbacks (for example).

    Attributes:
        proposals: Queue of outstanding proposals
        num_eval_pending: Number of pending evaluations
        num_eval_running: Number of running evaluations
        num_eval_outstanding: Number of outstanding evaluations
    """

    def __init__(self):
        self.proposals = deque([])
        self.num_eval_pending = 0
        self.num_eval_running = 0
        self._in_flight = {}

    @property
    def num_eval_outstanding(self):
        "Number of outstanding function evaluations."
        return self.num_eval_pending + self.num_eval_running

    def put(self, proposal):
        "Put a non-retry proposal in the queue."
        self.proposals.append(proposal)

    def rput(self, proposal):
        "Put a retry proposal in the queue."
        if not hasattr(proposal, 'retry'):
            logger.debug("Setting up retry")
            self._set(proposal, proposal.copy())
            self.retry = True
            if proposal.action == 'eval':
                proposal.add_callback(self.on_reply)
                self.num_eval_pending += 1
            elif proposal.action == 'kill':
                proposal.add_callback(self.on_kill_reply)
            elif proposal.action == 'terminate':
                proposal.add_callback(self.on_terminate_reply)
        else:
            logger.debug("Skip retry setup -- already under management")
        self.put(proposal)

    def get(self):
        "Pop a proposal from the queue."
        if self.proposals:
            return self.proposals.popleft()

    def empty(self):
        "Check if the queue is empty"
        return not self.proposals

    def on_reply_accept(self, proposal):
        "Process accepted eval+retry proposal"
        logger.debug("Recording accepted eval+retry proposal")
        self.num_eval_pending -= 1
        self.num_eval_running += 1
        self._rekey(proposal, proposal.record)

    def on_reply_reject(self, proposal):
        "Resubmit rejected eval+retry proposal"
        self.num_eval_pending -= 1
        self._resubmit(proposal)

    def on_kill_reply_reject(self, proposal):
        "Resubmit rejected kill+retry proposal"
        self._resubmit(proposal)

    def on_terminate_reply_reject(self, proposal):
        "Resubmit rejected termination+retry proposal"
        self._resubmit(proposal)

    def on_complete(self, record):
        "Clean up after completed eval+retry"
        logger.debug("Clean up after complete eval+retry")
        self.num_eval_running -= 1
        self._pop(record)

    def on_kill(self, record):
        "Resubmit proposal for killed or cancelled eval+retry"
        self.num_eval_running -= 1
        self._resubmit(record)

    def _set(self, key, proposal):
        "Record a retry proposal draft associated with a given key."
        self._in_flight[id(key)] = proposal

    def _pop(self, key):
        "Pop a retry proposal draft from the in-flight dictionary"
        return self._in_flight.pop(id(key))

    def _rekey(self, old_key, new_key):
        "Change the key on a retry proposal draft."
        logger.debug("Rekey retry proposal")
        self._set(new_key, self._pop(old_key))

    def _resubmit(self, key):
        "Recycle a previously-submitted retry proposal."
        logger.debug("Resubmitting retry proposal")
        self.rput(self._pop(key))


class FixedSampleStrategy(BaseStrategy):
    """Sample at a fixed set of points.

    The fixed sampling strategy is appropriate for any non-adaptive
    sampling scheme.  Since the strategy is non-adaptive, we can
    employ as many available workers as we have points to process.  We
    keep trying any evaluation that fails, and suggest termination
    only when all evaluations are complete.  The points in the
    experimental design can be provided as any iterable object (e.g. a
    list or a generator function).  One can use a generator for an
    infinite sequence if the fixed sampling strategy is used in
    combination with a strategy that provides a termination criterion.
    """

    def __init__(self, points):
        """Initialize the sampling scheme.

        Args:
            points: Points list or generator function.
        """
        def point_generator():
            "Generator wrapping the points list."
            for point in points:
                yield point
        self.point_generator = point_generator()
        self.retry = RetryStrategy()

    def propose_action(self):
        "Propose an action based on outstanding points."
        try:
            if self.retry.empty():
                logger.debug("Getting new point from fixed schedule")
                point = next(self.point_generator)
                proposal = self.propose_eval(point)
                self.retry.rput(proposal)
            return self.retry.get()
        except StopIteration:
            logger.debug("Done fixed schedule; {0} outstanding evals".format(
                self.retry.num_eval_outstanding))
            if self.retry.num_eval_outstanding == 0:
                return self.propose_terminate()


class CoroutineStrategy(BaseStrategy):
    """Event-driven to serial adapter using coroutines.

    The coroutine strategy runs a standard sequential optimization algorithm
    in a Python coroutine, with which the strategy communicates via send/yield
    directives.  The optimization coroutine yields the parameters at which
    it would like function values; the strategy then requests the function
    evaluation and returns the value with a send command when it becomes
    available.

    Attributes:
        coroutine: Optimizer coroutine
        rvalue:    Function to map records to values
        retry:     Retry manager
    """

    def __init__(self, coroutine, rvalue=lambda r: r.value):
        "Initialize the strategy."
        self.coroutine = coroutine
        self.rvalue = rvalue
        self.retry = RetryStrategy()
        try:
            logger.debug("Get first point proposal from coroutine")
            self.retry.rput(self.propose_eval(next(self.coroutine)))
        except StopIteration:
            logger.debug("Coroutine returned without yield")
            self.retry.rput(self.propose_terminate())

    def propose_action(self):
        "If we have a pending request, propose it."
        if not self.retry.empty():
            return self.retry.get()

    def on_complete(self, record):
        "Return point to coroutine"
        try:
            logger.debug("Return value to coroutine")
            params = self.coroutine.send(self.rvalue(record))
            logger.debug("Set up next point proposal from coroutine")
            self.retry.rput(self.propose_eval(params))
        except StopIteration:
            logger.debug("Coroutine returned without yield")
            self.retry.rput(self.propose_terminate())


class CoroutineBatchStrategy(BaseStrategy):
    """Event-driven to synchronous parallel adapter using coroutines.

    The coroutine strategy runs a synchronous parallel optimization
    algorithm in a Python coroutine, with which the strategy
    communicates via send/yield directives.  The optimization
    coroutine yields batches of parameters (as lists) at which it
    would like function values; the strategy then requests the
    function evaluation and returns the records associated with those
    function evaluations when all have been completed.

    NB: Within a given batch, function evaluations are attempted in
    the reverse of the order in which they are specified.  This should
    make no difference for most things (the batch is not assumed to have
    any specific order), but it is significant for testing.

    Attributes:
        coroutine: Optimizer coroutine
        retry:     Retry strategy for in-flight actions
        results:   Completion records to be returned to the coroutine
    """

    def __init__(self, coroutine, rvalue=lambda r: r.value):
        self.coroutine = coroutine
        self.rvalue = rvalue
        self.retry = RetryStrategy()
        self.results = []
        try:
            logger.debug("Get first proposed batch from coroutine")
            self.start_batch(next(self.coroutine))
        except StopIteration:
            pass

    def start_batch(self, xs):
        "Start evaluation of a batch of points."
        for ii in range(len(xs)):
            proposal = self.propose_eval(xs[ii])
            proposal.batch_id = ii
            self.retry.rput(proposal)
        self.results = [None for x in xs]

    def propose_action(self):
        "If we have a pending request, propose it."
        try:
            if (self.retry.empty() and self.retry.num_eval_outstanding == 0):
                logger.debug("Return batch to coroutine")
                batch = self.coroutine.send(self.results)
                logger.debug("Queue up next batch from coroutine")
                self.start_batch(batch)
            if not self.retry.empty():
                return self.retry.get()
        except StopIteration:
            logger.debug("Coroutine returned without yield")
            return self.propose_terminate()

    def on_reply_accept(self, proposal):
        "If proposal accepted, wait for result."
        proposal.record.batch_id = proposal.batch_id

    def on_complete(self, record):
        "Return or re-request on completion or cancellation."
        logger.debug("Got batch value ({0}/{1})".format(
            record.batch_id, len(self.results)))
        self.results[record.batch_id] = self.rvalue(record)


class RunTerminatedException(Exception):
    pass


class PromiseStrategy(BaseStrategy):
    """Provides a promise-based asynchronous evaluation interface.

    A promise (aka a future) is a common concurrent programming abstraction.
    A caller requests an asynchronous evaluation, and the call returns
    immediately with a promise object.  The caller can check the promise
    object to see if it has a value ready, or wait for the value to become
    ready.  The callee can set the value when it is ready.

    Attributes:
        proposalq: Queue of proposed actions caused by optimizer.
        valueq:    Queue of function values from proposed actions.
        rvalue:    Function to extract return value from a record
        block:     Flag whether to block on proposal pop
    """

    class Promise(object):
        """Evaluation promise.

        Properties:
            value: Promised value.  Block on read if not ready.
        """

        def __init__(self, valueq):
            self._value = None
            self.valueq = valueq

        def ready(self):
            "Check whether the value is ready (at consumer)"
            logger.debug("Checking if promise ready")
            while not self.valueq.empty():
                msg = self.valueq.get()
                msg()
            logger.debug("Promise ready = {0}".format(self._value is not None))
            return self._value is not None

        @property
        def value(self):
            "Wait on the value (at consumer)"
            logger.debug("Waiting on promise")
            while self._value is None:
                msg = self.valueq.get()
                msg()
            logger.debug("Promise ready")
            return self._value

        @value.setter
        def value(self, fx):
            "Set the value (at producer)"
            logger.debug("Sending value to promise object")
            self.valueq.put(lambda: self._set(fx))

        def _set(self, fx):
            "Set the value (at the consumer)"
            self._value = fx

    # Controller-facing routines

    def __init__(self, rvalue=lambda r: r.value, block=True):
        "Initialize the strategy."
        self.proposalq = Queue.Queue()
        self.valueq = Queue.Queue()
        self.rvalue = rvalue
        self.block = block

    def propose_action(self):
        "Provide proposals from the queue."
        if self.block or not self.proposalq.empty():
            return self.proposalq.get()

    def on_reply_accept(self, proposal):
        "Make sure we copy over the promise to the feval record."
        proposal.record.promise = proposal.promise

    def on_reply_reject(self, proposal):
        "Re-submit the proposal with the same promise."
        new_proposal = self.propose_eval(*proposal.args)
        new_proposal.promise = proposal.promise
        self.proposalq.put(new_proposal)

    def on_terminate_reply_reject(self, proposal):
        "Re-submit the termination proposal."
        self.proposalq.put(self.propose_terminate())

    def on_complete(self, record):
        "Send the value to the consumer via the promise."
        record.promise.value = self.rvalue(record)

    def on_kill(self, record):
        "Re-submit the proposal with the same promise."
        proposal = self.propose_eval(*record.params)
        proposal.promise = record.promise
        self.proposalq.put(proposal)

    def on_terminate(self):
        "Throw an exception at the consumer if still running on termination"
        self.valueq.put(self._throw_terminate)

    # Client-facing routines

    def promise_eval(self, *args):
        "Request a function evaluation and return a promise object."
        logger.debug("Returning promise value")
        proposal = self.propose_eval(*args)
        proposal.promise = self.Promise(self.valueq)
        self.proposalq.put(proposal)
        return proposal.promise

    def blocking_eval(self, *args):
        "Request a function evaluation and block until done."
        return self.promise_eval(*args).value

    def blocking_evals(self, xs):
        "Request a list of function evaluations."
        promises = [self.promise_eval(x) for x in xs]
        return [p.value for p in promises]

    def terminate(self):
        "Request termination."
        logger.debug("PromiseStrategy queueing termination request")
        self.proposalq.put(self.propose_terminate())

    def _throw_terminate(self):
        logger.debug("Stop worker with RunTerminatedException")
        raise RunTerminatedException()


class ThreadStrategy(PromiseStrategy):
    """Event-driven to serial adapter using threads.

    The thread strategy runs a standard sequential optimization algorithm
    in a separate thread of control, with which the strategy communicates
    via promises.  The optimizer thread intercepts function evaluation requests
    and completion and turns them into proposals for the strategy, which
    it places in a proposal queue.  The optimizer then waits for the requested
    values (if any) to appear in the reply queue.

    Attributes:
        optimizer: Optimizer function (takes objective as an argument)
        proposalq: Queue of proposed actions caused by optimizer.
        valueq:    Queue of function values from proposed actions.
        thread:    Thread in which the optimizer runs.
        proposal:  Proposal that the strategy is currently requesting.
        rvalue:    Function mapping records to values
    """

    class OptimizerThread(threading.Thread):
        def __init__(self, strategy, optimizer):
            threading.Thread.__init__(self)
            self.strategy = strategy
            self.optimizer = optimizer

        def run(self):
            try:
                self.optimizer(self.strategy)
            except RunTerminatedException:
                logger.debug("Caught RunTerminatedException in opt thread")
            finally:
                self.strategy.terminate()

    def __init__(self, controller, optimizer, rvalue=lambda r: r.value):
        logger.debug("Initialize ThreadStrategy")
        PromiseStrategy.__init__(self, rvalue, block=False)
        self.thread = self.OptimizerThread(self, optimizer)
        self.thread.start()
        controller.add_term_callback(self.on_terminate)

    def on_terminate(self):
        logger.debug("In ThreadStrategy on_terminate")
        PromiseStrategy.on_terminate(self)
        logger.debug("ThreadStrategy waiting on worker thread join")
        self.thread.join()


class CheckWorkerStrategy(object):
    """Preemptively kill eval proposals when there are no workers.

    A strategy like the fixed sampler is simple-minded, and will
    propose a function evaluation even if the controller has no
    workers available to carry it out.  This wrapper strategy
    intercepts proposals from a wrapped strategy like the fixed
    sampler, and only submits evaluation proposals if workers are
    available.
    """

    def __init__(self, controller, strategy):
        "Initialize checker strategy."
        self.controller = controller
        self.strategy = strategy

    def propose_action(self):
        "Generate filtered action proposal."
        proposal = self.strategy.propose_action()
        if (proposal and proposal.action == 'eval' and
                not self.controller.can_work()):
            logger.debug("Reject eval proposal (no worker available)")
            proposal.reject()
            return None
        return proposal


class SimpleMergedStrategy(object):
    """Merge several strategies by taking the first valid proposal from them.

    The simplest combination strategy is to keep a list of several
    possible strategies in some priority order, query them for
    proposals in priority order, and pass on the first plausible
    proposal to the controller.

    Attributes:
        controller: Controller object used to determine whether we can eval
        strategies: Prioritized list of strategies
    """

    def __init__(self, controller, strategies):
        "Initialize merged strategy."
        self.controller = controller
        self.strategies = strategies

    def propose_action(self):
        "Go through strategies in order and choose the first valid action."
        for strategy in self.strategies:
            proposal = strategy.propose_action()
            if proposal is None:
                pass
            elif proposal.action == 'eval' and not self.controller.can_work():
                logger.debug("Reject eval proposal (no worker available)")
                proposal.reject()
            else:
                return proposal


class MultiStartStrategy(object):
    """Merge several strategies by taking the first valid eval proposal.

    This strategy is similar to the SimpleMergedStrategy, except that we
    only terminate once all the worker strategies have voted to terminate.

    Attributes:
        controller: Controller object used to determine whether we can eval
        strategies: Prioritized list of strategies
    """

    def __init__(self, controller, strategies):
        "Initialize merged strategy."
        self.controller = controller
        self.strategies = strategies

    def propose_action(self):
        """Go through strategies in order and choose the first valid action.
        Terminate iff all workers vote to terminate.
        """
        proposals = [strategy.propose_action() for strategy in self.strategies]
        proposals.reverse()
        chosen_proposal = None
        terminate_votes = 0
        for proposal in proposals:
            if not proposal:
                pass
            elif proposal.action == 'eval' and self.controller.can_work():
                chosen_proposal = proposal
            elif proposal.action == 'kill':
                chosen_proposal = proposal
            elif proposal.action == 'terminate':
                terminate_votes += 1
        for proposal in proposals:
            if proposal is not None and proposal != chosen_proposal:
                proposal.reject()
        if terminate_votes == len(proposals):
            return Proposal("terminate")
        return chosen_proposal


class MaxEvalStrategy(object):
    """Recommend termination of the iteration after some number of evals.

    Recommends termination after observing some number of *completed*
    function evaluations.  We allow more than the requisite number
    of evaluations to be started, since there's always the possibility
    that one of our outstanding evaluations won't finish.

    Attributes:
        counter: Number of completed evals
        max_counter: Max number of evals
    """

    def __init__(self, controller, max_counter):
        """Initialize the strategy.

        Args:
            controller: Used for registering the feval callback
            max_counter: Maximum number of evals desired
        """
        self.counter = 0
        self.max_counter = max_counter
        controller.add_feval_callback(self.on_new_feval)

    def on_new_feval(self, record):
        "On every feval, add a callback."
        record.add_callback(self.on_update)

    def on_update(self, record):
        "On every completion, increment the counter."
        if record.is_completed:
            self.counter += 1
        logger.debug("MaxEvalStrategy: completed {0}/{1}".format(
            self.counter, self.max_counter))

    def propose_action(self):
        "Propose termination once the eval counter is high enough."
        if self.counter >= self.max_counter:
            logger.debug("MaxEvalStrategy: request termination")
            return Proposal('terminate')
        else:
            return None


class InputStrategy(BaseStrategy):
    """Insert requests from the outside world (e.g. from a GUI)."""

    def __init__(self, controller, strategy):
        self.controller = controller
        self.strategy = strategy
        self.retry = RetryStrategy()

    def _propose(self, proposal, retry):
        if retry:
            self.retry.rput(proposal)
        else:
            self.retry.put(proposal)
        self.controller.ping()

    def eval(self, params, retry=True):
        "Request a new function evaluation"
        self._propose(self.propose_eval(params), retry)

    def kill(self, record, retry=False):
        "Request a function evaluation be killed"
        self._propose(self.propose_kill(record), retry)

    def terminate(self, retry=True):
        "Request termination of the optimization"
        self._propose(self.propose_terminate(), retry)

    def propose_action(self):
        if not self.retry.empty():
            return self.retry.get()
        return self.decorate_proposal(self.strategy.propose_action())


class AddArgStrategy(BaseStrategy):
    """Add extra arguments to evaluation proposals.

    Decorates evaluation proposals with extra arguments.  These may
    reflect termination criteria, advice to workers about which of
    several methods to use, etc.  The extra arguments can be static
    (via the extra_args attribute), or they may be returned from the
    get_extra_args function, which child classes can override.
    The strategy *overwrites* any existing list of extra arguments,
    so it is probably a mistake to compose strategies in a way that
    involves multiple objects of this type.

    Attributes:
        strategy: Parent strategy
        extra_args: Extra arguments to be added
    """

    def __init__(self, strategy, extra_args=None):
        self.strategy = strategy
        self.extra_args = extra_args

    def add_extra_args(self, record):
        """Add any extra arguments to an eval record.

        Args:
            record: Record to be decorated
        """
        record.extra_args = self.extra_args

    def on_reply_accept(self, proposal):
        self.add_extra_args(proposal.record)

    def propose_action(self):
        return self.decorate_proposal(self.strategy.propose_action())


class ChaosMonkeyStrategy(object):
    """Randomly kill running function evaluations.

    The ChaosMonkeyStrategy kills function evaluations at random from
    among all active function evaluations.  Attacks are associated
    with a Poisson process where the mean time between failures is
    specified at startup.  Useful mostly for testing the resilience of
    strategies to premature termination of their evaluations.  The
    controller may decide to ignore the proposed kill actions, in
    which case the monkey's attacks are ultimately futile.
    """

    def __init__(self, controller, strategy, mtbf=1):
        """Release the chaos monkey!

        Args:
            controller: The controller whose fevals we will kill
            strategy: Parent strategy
            mtbf: Mean time between failure (assume Poisson process)
        """
        self.controller = controller
        self.strategy = strategy
        self.running_fevals = []
        self.lam = 1.0/mtbf
        self.target = None
        controller.add_feval_callback(self.on_new_feval)
        controller.add_timer(random.expovariate(self.lam), self.on_timer)

    def on_new_feval(self, record):
        "On every feval, add a callback."
        record.chaos_target = False
        record.add_callback(self.on_update)

    def on_update(self, record):
        "On every completion, remove from list"
        if record.is_running and not record.chaos_target:
            record.chaos_target = True
            self.running_fevals.append(record)
        elif record.is_done and record.chaos_target:
            record.chaos_target = False
            self.running_fevals.remove(record)

    def on_timer(self):
        if self.running_fevals:
            record = self.running_fevals.pop()
            record.chaos_target = False
            self.target = record
            self.controller.ping()
        self.controller.add_timer(random.expovariate(self.lam), self.on_timer)

    def propose_action(self):
        if self.target:
            logger.info("Monkey attack! {0}".format(self.target.params))
            proposal = Proposal('kill', self.target)
            self.target = None
        else:
            proposal = self.strategy.propose_action()
        return proposal



"""
Script to update correctly the main GUI (.py) file from the Qt design (.ui) file
"""
import os
import sys
from subprocess import call

if __name__ == '__main__':
    # pyrcc5 icons.qrc -o icons_rc.py
    # pyuic5 -x MainWindow.ui -o MainWindow.py

    py = sys.executable
    folder = os.path.dirname(py)
    f1 = folder.split(os.sep)[-1]

    if f1 == 'bin':
        fbase = folder
        script_name = 'pyside2-uic'
    else:
        fbase = os.path.join(folder, 'Scripts')
        script_name = 'pyside2-uic-script.py'

    # update icon/images resources
    # call([py, os.path.join(fbase, 'pyside2-rcc'), 'icons.qrc', '-o', 'icons_rc.py'])

    #  update files
    file_names_ui = ['gui.ui']
    file_names = ['gui.py']

    for filename, filename_ui in zip(file_names, file_names_ui):

        # update ui handler file
        call([py, os.path.join(fbase, script_name), '-x', filename_ui, '-o', filename])

        # replace annoying text import
        # Read in the file
        with open(filename, 'r') as file:
            file_data = file.read()

        # Replace the target string
        file_data = file_data.replace('import icons_rc', 'from .icons_rc import *')

        # Write the file out again
        with open(filename, 'w') as file:
            file.write(file_data)
